# -*- coding: utf-8 -*-

import Spr
from time import sleep
from threading import Thread
import copy

from .Utility import *
from .Database import ValueSet

from .Planner import GetSatisfactionRatio

class ActionController(Thread):
	#status
	ST_STOP,ST_RUN,ST_IDLE,ST_DISPOSE = range(4)
	
	@classmethod
	def CreateFromCreature( cls , cr ):
		self = cls()
		Thread.__init__(self)
		self.setDaemon(True)
		
		self.cr = cr
		
		self.plan = None
		self.planBuff = None
		
		self.planAborted = False 
		
		self.currentAction = 0
		self.status = self.ST_IDLE
		Thread.start(self)
		return self
	
	def run(self):
		while( not self.status == self.ST_DISPOSE ):
		
			if self.planAborted:
				self.plan = None
				self.currentAction = 0
				self.planAborted = False
			
			if self.planBuff:
				self.plan = self.planBuff
				self.planBuff = None
				self.currentAction = 0
			
			if (self.status == self.ST_RUN) and self.plan:
				# 正常にプランが完了したとき
				if self.currentAction >= len(self.plan):
					dprint("done")
					self.status = self.ST_IDLE
					self.currentAction = 0
					self.plan = None
					continue
				
				
				act,env  = self.plan[self.currentAction]
				
				# 前提条件が満たされていない場合の処理
				ratio = GetSatisfactionRatio(self.cr.models["current"],act.Precond())
				if ratio < 1.0:
					dprint("プラン破棄 ratio=",ratio,"env=",env)
					self.AbortPlan()
					continue
				
				# 現在のアクションを実行
				act.Do(env)
				
				# 行動が失敗していた時の処理（イベントにする？
				#if "ACT_STATUS" in env:
				#	if env["ACT_STATUS"] == "Aborted":
				#		self.AbortPlan()
				
				
				# 終了条件を満たしていたら次の行動へ
				if "ACT_STATUS" in  env:
					dprint("アクション完了 ",act,"env=",env)
					if env["ACT_STATUS"] == "Compleated":
						self.currentAction += 1
			
			else:
				self.status = self.ST_IDLE
			sleep(0.3)
	
	def Dispose(self):
		self.status = self.ST_DISPOSE
	
	def SetPlan(self,plan):
		self.planBuff = plan
	
	def GetPlan(self):
		return self.plan
	
	def AbortPlan(self):
		self.planAborted = True
		self.Stop()
	
	def Stop(self):
		self.status = self.ST_STOP
	
	def Join(self):
		self.Stop()
		while not self.status == self.ST_IDLE:
			sleep(0.01)
		
	def Start(self):
		self.status = self.ST_RUN
		
	def GetCurrentAction(self):
		return self.plan[self.currentAction]
	
	def IsEmpty(self):
		return ( self.plan == None and self.planBuff == None)
	
	# base隠蔽
	def start(self):
		self.Start()
	
	def join(self):
		self.Join()
	
class Action:
	def __init__(self,name,cr,owner):
		self.name = name
		self.cr = cr
		self.owner = owner
	
	def __repr__(self):
		return self.owner.name + ".ACT<" + self.name + ">"
	
	def __str__(self):
		return self.owner.name + ".ACT<" + self.name + ">"
	
	def __deepcopy__(self,dict):
		ret = self.__new__(type(self),"",None)
		ret.name = copy.deepcopy(self.name)
		ret.cr   = copy.copy(self.cr)
		ret.owner = self.owner
		return ret
	
	def Precond(self):
		pass
	
	def Effect(self,env):
		return env.copy()
	
	def EnumFulfilled(self):
		pass
	
	def Do(self,env):
		dprint("doing '",self.name,"'")
	
	def Init(self):
		pass
	
	def Terminate(self):
		pass

class ActionLook(Action):
	def __init__(self,cr,owner):
		Action.__init__(self,"Look",cr,owner)
	
	#もし成功したらどういう変化が起こるか
	#現在の黒板のコピーを編集して返す
	def Effect(self,current):
		after = current.copy()
		owner = self.owner.Clone()
		after[self.owner.phSolid] = owner
		
		if not owner : return after
		
		v,c = owner.GetAttr("position")
		if v:
			owner.SetAttr("position",ValueSet(v,1.0))
		
		return after
	
	
	#行動するためにさらに必要な条件があれば条件の辞書を返す
	#条件が関数の場合は、Act.Effectやmodelなどから渡された、
	#行動後や現在のモデルの変化をあらわすenvを引数として渡される
	def Precond(self):
		"""
		precond = {}
		#lso = self.cr.models["current"][self.owner.phSolid]
		lso = current[self.owner.phSolid]
		
		if lso:
			pos,cf = lso.GetAttr("position")
			print(pos,cf)
			if pos and cf > 0.95:
				pass
			else:
				precond["知っている"] = True
		else:
			precond["知っている"] = True
		return precond
		"""
		def c1gen(owner):
			def c1(env):
				nonlocal owner
				try:
					#effect内のownerの情報を取得
					owner = env[owner.phSolid]
					#場所について情報がない場合はFalse
					if owner.GetAttr("position").value == None: return False
					dprint("visible true")
					#すべてクリアした場合はTrue
					return True
				except:
					pass
					#dprint("some error")
				return False
			return c1
		return {"KnowPos":c1gen(self.owner)}
		
	def Do(self,env):
		owner = self.owner
		tarpos,cf = owner.GetAttr("position")
		self.cr.HeadController.LookAt(tarpos,Spr.Vec3d(),11)
		
		env["ACT_STATUS"] = "Compleated"
	
class ActionRehearsal(Action):
	def __init__(self,cr,owner):
		Action.__init__(self,"Rehearsal",cr,owner)
	
	def Effect(self,current):
		after = current.copy()
		owner = self.owner.Clone()
		after[self.owner.phSolid] = owner
		
		after["知っている"] = True
		return after
		
	def Precond(self):
		precond = {}
		precond["本気出す"] = True
		return precond

class ActionMove(Action):
	def __init__(self,cr,owner):
		Action.__init__(self,"Move",cr,owner)
	
	def Effect(self,current):
		after = current.copy()
		owner = self.owner.Clone()
		after[self.owner.phSolid] = owner
		
		v,c = owner.GetAttr("position")
		if v: 
			after["BasePos"] = v
		return after
		
	def Precond(self):
		return {}
	
	def Do(self,env):
		v,c = self.owner.GetAttr("position")
		if v:
			self.cr.MoveController.MoveTo(v)
			
			basePos = self.cr.Body.FindByLabel("Base").GetPHSolid().GetFramePosition()
		
			dist = (basePos- v).norm()
			if dist < 5:
				env["ACT_STATUS"] = "Compleated"
			

class ActionReachRH(Action):
	def __init__(self,cr,owner):
		Action.__init__(self,"ReachRH",cr,owner)
	

	def Effect(self,current):
		after = current.copy()
		owner = self.owner.Clone()
		after[self.owner.phSolid] = owner
		
		owner.SetAttr("触っている", True)
		
		return after
	
	def Do(self,env):
		v,c = self.owner.GetAttr("position")
		if v:
			self.cr.RightHand.SetTargetPos(v)
			self.cr.RightHand.Start()
		dist = (self.cr.RightHand.GetPHSolid().GetFramePosition()-v).norm()
		print("++++++++++++++++++++++",dist)
		if (dist<1.5):
			env["ACT_STATUS"] = "Compleated"
		
	
	def Precond(self):
		def c1gen(owner):
			def c1(env):
				nonlocal owner
				try:
					#effect内のownerの情報を取得
					owner = env[owner.phSolid]
					#場所について情報がない場合はFalse
					if owner.GetAttr("position").value == None: return False
					if owner.GetAttr("position").conf < 0.95: return False
					dprint("visible true")
					#すべてクリアした場合はTrue
					return True
				except:
					pass
					#dprint("some error")
				return False
			return c1
		
		def c2gen(owner):
			def c2(env):
				nonlocal owner
				print(type(env))
				#effect内のownerの情報を取得
				owner = env[owner.phSolid]
				#遠い場合はFalse
				basePos = None
				if not "BasePos" in env:
					basePos = self.cr.Body.FindByLabel("Base").GetPHSolid().GetFramePosition()
				else:
					basePos = env["BasePos"]
				
				dist = (owner.GetAttr("position").value - basePos).norm()
				if (dist> 5):
					dprint("reachable false. dist=",dist)
					return False
				dprint("reachable true")
				#すべてクリアした場合はTrue
				return True
				return False
			return c2
		
		return {"CKnowPos":c1gen(self.owner),"CReachable":c2gen(self.owner)}

"""
#for プランニングテスト
class TA1(Action):
	def __init__(self,cr):
		Action.__init__(self,"Test1",cr)
	
	def Effect(self,current):
		return {"T1":True}
	
	def Precond(self):
		def f(env):
			if "T3" in env and env["T3"] == True:
				return True
			if "T4" in env and env["T4"] == True:
				return True
			return False
		return {"condT1":f}
	

class TA2(Action):
	def __init__(self,cr):
		Action.__init__(self,"Test2",cr)
	
	def Effect(self,current):
		return {"T2":True}
	
	def Precond(self):
		def f1(env):
			if "T5" in env and env["T5"] == True:
				return True
		def f2(env):
			if "T6" in env and env["T6"] == True:
				return True
		return {"condT2a":f1,"condT2b":f2}


class TA3(Action):
	def __init__(self,cr):
		Action.__init__(self,"Test3",cr)
	
	def Effect(self,current):
		return {"T3":True}
	
	def Precond(self):
		return {}


class TA4(Action):
	def __init__(self,cr):
		Action.__init__(self,"Test4",cr)
	
	def Effect(self,current):
		return {"T4":True}
	
	def Precond(self):
		return {}

class TA5(Action):
	def __init__(self,cr):
		Action.__init__(self,"Test5",cr)
	
	def Effect(self,current):
		return {"T5":True}
	
	def Precond(self):
		return {}

class TA6(Action):
	def __init__(self,cr):
		Action.__init__(self,"Test6",cr)
	
	def Effect(self,current):
		return {"T6":True}
	
	def Precond(self):
		return {}

"""
