# -*- coding: utf-8 -*-

import Spr
from time import sleep
from threading import Thread

from .LSolid import *
from .Model import *
from .Database import *
# from .Planner import *
from .Controller import *
# from .Action import *

def rad(deg):
	return deg / 180.0 * 3.1415926

class TouchInfo:
	def __init__(self,position,force,area,phSolid,bodySolid):
		self.position = position
		self.force = force
		self.area = area
		self.phSolid = phSolid
		self.bodySolid = bodySolid
		self.pressure = self.force.norm() / (self.area if self.area != 0 else 1)
		
class TouchSensor:
	def __init__(self,cr):
		self.cr = cr
		
	def GetTouchInfos(self):
		touchInfos = []
		# bodySo ->ボディの一部 cotnactSo ->接触している物
		for bodySo in self.cr.Body.GetSolids():
			for i,contactSo in enumerate(bodySo.GetContactSolids()):
				phSolid = contactSo
				position = bodySo.GetContactPosition(i)
				force = bodySo.GetContactForce(i)
				area = bodySo.GetContactArea(i)
				touchInfos.append(TouchInfo( position,force,area,phSolid,bodySo ))
		return touchInfos

class Creature(Thread):
	@classmethod
	def CreateFromCRCreature( cls,cr,phSdk,attrDB,worldModel, debugOut = None , phSceneMemory = None):
		self = cls()
		Thread.__init__(self)
		self.setDaemon(True)
		
		self.CR = cr
		self.attrDB = attrDB
		self.worldModel = worldModel
		
		# body
		if cr.NBodies() < 1:
			raise Exception("No Body exists!!")
		self.Body = cr.GetBody(0)
		self.LeftHand 	= self.Body.FindByLabel("LeftHand")
		self.RightHand 	= self.Body.FindByLabel("RightHand")
		self.Head		= self.Body.FindByLabel("Head")
		self.Waist		= self.Body.FindByLabel("Waist")
		self.Base 		= self.Body.FindByLabel("Base")
		
		# controller
		if cr.NEngines() < 1:
			raise Exception("No NeckController exists!!")
		self.HeadController = cr.GetEngines()[0]
		#self.MoveController = MoveController( self.Base.GetPHSolid() )
		
		# sensor
		self.VisualSensor = self.Body.FindByLabel("Head")
		self.VisualSensor.SetViewArea(Spr.Posed(0.710107,-0.710107,0,0, 0,0,0), rad(120), rad(90))
		self.TouchSensor = TouchSensor(self)
		
		
		self.DebugOut	= debugOut
		
		# sofloorは例外処理をするため、phSolidを取得
		soFloor = self.Body.GetScene().GetPHScene().FindObject("soFloor")
		
		self.models = {}
		model = CreatureModel("CurrentModel",self,phSdk)
		model.RegisterKS(KSBase(model.GetInternalPHScene()))
		model.RegisterKS(KSPercept(
									self.VisualSensor,
									self.TouchSensor,
									self.worldModel,
									[ikSo.GetPHSolid() for ikSo in self.Body.GetSolids()] + [soFloor] #bodyの要素は除外
								)
						)
		model.RegisterKS(KSAttention())
		model.RegisterKS(KSAttribute(self,self.attrDB))
		model.RegisterKS(KSPredict())
		self.models["current"]  = model
		
		self.rules = {}
		
		return self
	
	def ModifySpringDamper(self, K = None , D = None):
		# spring damper
		springRatio = K if K else 1
		damperRatio = D if D else 1
		for jo in self.Body.GetJoints():
			jo.GetIKActuator().SetSpring( jo.GetIKActuator().GetSpring() * springRatio )
			jo.GetIKActuator().SetDamper( jo.GetIKActuator().GetDamper() * damperRatio )
			jo.GetPHJoint().SetSpring( jo.GetPHJoint().GetSpring() * springRatio )
			jo.GetPHJoint().SetDamper( jo.GetPHJoint().GetDamper() * damperRatio )
	
	# Creatureのループ
	def run(self):
		while True:
			# KnowledgeSourceのチェーンを呼び出してModelを更新する
			for key,model in self.models.items():
				model.Step()
			
			# Ruleのチェーンを呼び出して適用する
			for key,rule in sorted(self.rules.items()):
				exec(rule);
			
			#debug出力obj
			sleep(0.03)
