# -*- coding: utf-8 -*-

import Spr
import re
import copy

from collections import namedtuple

class ValueSet (namedtuple("ValSet","value, conf")):
	def __init__(self,val,conf):
		self.oldnew = self.__new__
		self.__new__ = lambda cls,v,c : cls.oldnew(cls,v, (c if c > 0.001 else 0) )
		
		super(ValueSet,self).__init__(val,conf)
		
	def __str__(self):
		if type(self.value) == float:
			return "ValueSet(%.3f,%.3f)" % (self.value,self.conf)
		else:
			return "ValueSet(%r,%.3f)" % (self.value,self.conf)
	
# Noneだといろいろ面倒なので、エラーにならないオブジェクト
# DatabaseItem用
class EmptyObject:
	#鏡像（lhsとrhsが逆になること）にならないらしい。rhsが必ず比較対象
	def __eq__(lhs,rhs):
		if (rhs == "" or rhs == None):
			return True
		else:
			return False
	
	def __ne__(lhs,rhs):
		return not (lhs == rhs)
		
	def __len__(self):
		return 0
		
	def __call__(self,*arg):
		return None
	
	def __contains__(self,item):
		return False
		
	def __iter__(self):
		return iter([])

class DatabaseItem():
	def __init__(self):
		self.children = []
		self.parents = []
		self.value = {}
		self.name = ""
	
	def __contains__(self,key):
		if key in self.value:
			return True
		else:
			for parent in self.parents:
				if key in parent:
					return True
		return False
		
	def __getitem__(self,key):
		if key in self.value:
			return self.value[key]
		else:
			for parent in self.parents:
				return parent.__getitem__(key)
		
		#例外を発生させない
		return EmptyObject() #raise KeyError
	
	def __setitem__(self,key,val):
		if (val == None) and (key in  self.value):
			del self.value[key]
		else:
			self.value[key] = val
	
	def __delitem__(self,key):
		if key in self.value:
			del self.value[key]
	
	def __deepcopy__(self,dict):#おそらく不完全
		ret = DatabaseItem()
		ret.children = []
		ret.parents = self.parents
		ret.value = self.value.copy()
		ret.name = copy.deepcopy(self.name)
		return ret
		
	def __repr__(self):
		ret = self.ToDict()
		return "DatabaseItem (" + self.name + "): " + str(ret)
		
	def __str__(self):
		return self.__repr__()
	
	def DoChildren(self,func,**kwds):
		for child in self.children:
			func( child, **kwds )
			child.DoChildren( func, **kwds )
	
	def DoParents(self,func, **kwds):
		for parent in self.parents:
			parent.DoParents( func, **kwds )
			func( parent, **kwds )
	
	def ToDict(self,recursive = True):
		if not recursive:
			return self.value
		
		def EnumItemsR(self,**kwds):
			kwds["ret"].update( self.value )
		
		_ret = {}
		self.DoParents(EnumItemsR,ret=_ret)
		EnumItemsR(self,ret=_ret)
		return _ret
	
	#def Keys(self):
	#	def GetKeysR(self,**kwds):
	#		kwds["ret"].append([ k for k in self.value.keys()])
	#	
	#	_ret={}
	#	self.DoParents(GetKeysR,ret=_ret)
	#	GetKeysR(self,ret=_ret)
	#	return _ret
	
class Attribute(DatabaseItem):
	def __init__(self):
		self.children = []
		self.parents = []
		self.value = {}
		self.name = ""
		
		self.precondition = ""
		self.ratio = 0
		self.afordance = []

class Database:
	def __init__(self,name):
		self.name = name
		self.items = {}
		self.root = None

	class Error(Exception):
		def __init__(self, value):
			self.value = value
		def __str__(self):
			return repr(self.value)

	def BuildDB(self):
		#LoadFromFile(...)    #いずれ
		pass
		
	def PrintDB(self):
		def PrintR(self,**kwds):
			def PrintParents(parent,**kwds):
				if "ret" in kwds:
					kwds["ret"][0] = parent.name + "<-" + (kwds["ret"][0] if kwds["ret"][0] else "")
			bases = [""]
			self.DoParents(PrintParents,ret=bases)
			print(kwds["arg"] + bases[0] + self.name + " :: ",end='')
			if type(self.value) == dict:
				print(str(self.value.keys()))
			else:
				print(str(self.value))
		
		self.root.DoChildren(PrintR,arg="")
		

from .Action import *
class AttrDatabase(Database):
	def Interpret(self,lso,cr):
		val,cf = lso.GetAttr("taste")
		if val == "sweet":
			#食べ物グループ
			lso.SetAttr("Food",True)
		
		
		if True:
			# すべてのphSolidを持つ物体のグループ
			actions = []
			actions.append(ActionLook(cr,lso))
			#actions.append(ActionRehearsal(cr,lso))
			actions.append(ActionMove(cr,lso))
			actions.append(ActionReachRH(cr,lso))
			lso.SetAttr("afordance",actions)
			
			"""
			#for プランニングテスト
			actions = []
			actions.append( TA1(cr) )
			actions.append( TA2(cr) )
			actions.append( TA3(cr) )
			actions.append( TA4(cr) )
			actions.append( TA5(cr) )
			actions.append( TA6(cr) )
			lso.SetAttr("afordance",actions)
			"""
			
			
		
class WorldDatabase(Database):	
	def Interpret(self,lso):
		pat = re.compile("^([^0-9]+)[0-9]*$")  #ObjectName01 といった形で書くこと
		m = pat.match(lso.name)
		if m:
			key = m.groups()[0]
			if key in self.items:
				lso.__init__( lso.phSolid, self.items[key] )
			else:
				raise Database.Error(key + " is not in WorldDB")
			
		else:
			raise Database.Error(lso.name + " has name formatting error")
		
