# -*- coding: utf-8 -*-

import Spr
import copy

from .Database import ValueSet
from .LSolid import *

class KnowledgeSource:
	def __init__(self):
		pass
	
	def Step(self,model):
		pass

class KSBase(KnowledgeSource):
	def __init__(self,phScene):
		KnowledgeSource.__init__(self)
		self.stepCount = None
		self.phScene = phScene
	
	def Step(self,model):
		model.stepCount = self.phScene.GetCount()

class KSPercept(KnowledgeSource):
	def __init__(self,visualSensor,touchSensor,worldModel,exceptSolids):
		KnowledgeSource.__init__(self)
		self.visualSensor = visualSensor
		self.touchSensor = touchSensor
		self.worldModel = worldModel
		self.exceptSolids = exceptSolids
		
	def Step(self,model):
		# -- -- -- -- センサからの情報取得 -- -- -- --
		# from VisionSensor...
		visibleSolids = self.visualSensor.GetVisibleSolids()
		model.visibleSolids = visibleSolids
		model.eyePos = self.visualSensor.GetPHSolid().GetFramePosition()
		
		# from TouchSensor...
		touchInfos = self.touchSensor.GetTouchInfos()
		model.touchInfos = touchInfos
		
		# exceptSolidsをセンサ出力から除外
		for bodyPHSo in self.exceptSolids:
			if bodyPHSo in visibleSolids:
				visibleSolids.remove(bodyPHSo)
		for touchInfo in touchInfos:
			if touchInfo.phSolid in self.exceptSolids:
				touchInfos.remove(touchInfo)
		
		# last_view,last_touch属性を更新
		for so in visibleSolids:
			if not so in model:
				model[so] = LSolid(so)
			model[so].SetAttr("last_view",ValueSet(model.stepCount,1.0))
		
		for so in (touchInfo.phSolid for touchInfo in touchInfos):
			if not so in model:
				model[so] = LSolid(so)
			model[so].SetAttr("last_touch",ValueSet(model.stepCount,1.0))
		
		
		# -- -- -- -- センサ情報からLSolidをモデルに追加、更新 -- -- -- --
		# 視覚情報
		for lso in model.LSolids():
			lv,cf = lso.GetAttr("last_view")
			if lv == model.stepCount:
				lso.SetAttr("position",ValueSet(lso.phSolid.GetFramePosition,1.0))
				lso.SetAttr("velocity",ValueSet(lso.phSolid.GetVelocity,1.0))
				# worldModelから
				if lso in self.worldModel:
					val,cf = self.worldModel[lso].GetAttr("shape")
					if val:
						lso.SetAttr("shape",ValueSet(val,cf))
					val,cf = self.worldModel[lso].GetAttr("color")
					if val:
						lso.SetAttr("color",ValueSet(val,cf))
			# 見えてない場合はConfidenceを下げる
			elif not lv == None:
				# 前回見た時との時間差[s]
				dt = (model.stepCount - lv) * model.phScene.GetTimeStep()
				
				val,cf = lso.GetAttr("position")
				if val:
					# 1分で確信0
					lso.SetAttr("position",ValueSet(val, (1-dt/60 if dt < 60 else 0 ) ))
				
				val,cf = lso.GetAttr("velocity")
				if val:
					# 15秒で確信0
					lso.SetAttr("velocity",ValueSet(val, (1-dt/15 if dt < 15 else 0 ) ))
					
			
		# 触覚情報
		for lso in model.LSolids():
			lt,cf = lso.GetAttr("last_touch")
			if lt == model.stepCount:
				pos,cf = lso.GetAttr("position")
				if pos:
					#何度も呼ばれるからすぐ１になる・・・
					lso.SetAttr("position",ValueSet( lso.phSolid.GetFramePosition,(1.0 if cf + 0.02 > 1.0 else cf + 0.02) ))
				lso.SetAttr("mass",ValueSet(lso.phSolid.GetMass,0.7))
				# lsoの接触に関してのtouchInfoを取得
				touchInfos = [ touchInfo for touchInfo in model.touchInfos if touchInfo.phSolid == lso.phSolid ]
				lso.SetAttr("pressure",ValueSet(0,1))
				lso.SetAttr("touch_pair",ValueSet([],1))
				for touchInfo in touchInfos:
					pressure,cf = lso.GetAttr("pressure")
					lso.SetAttr("pressure",ValueSet(pressure+touchInfo.pressure,1))
					
					touch_pair,cf = lso.GetAttr("touch_pair")
					if touch_pair:
						lso.SetAttr("touch_pair",ValueSet(touch_pair.append(touchInfo.bodySolid),1))
					else:
						lso.SetAttr("touch_pair",ValueSet([touchInfo.bodySolid],1))
				
				# worldModelから
				if lso in self.worldModel:
					val,cf = self.worldModel[lso].GetAttr("shape")
					if val:
						lso.SetAttr("shape",ValueSet(val,cf))
			
		# 触っていない場合はConfidenceを下げる
			else:
				#val,cf = lso.GetAttr("position")
				#if val:
				#	cf *= 0.95
				#	lso.SetAttr("position",ValueSet(val,cf))
					
				val,cf = lso.GetAttr("mass")
				if val:
					cf *= 0.95
					lso.SetAttr("mass",ValueSet(val,cf))
				
				val,cf = lso.GetAttr("pressure")
				if val:
					cf *= 0.95
					lso.SetAttr("pressure",ValueSet(val,cf))
				
				val,cf = lso.GetAttr("touch_pair")
				if val:
					cf *= 0.95
					lso.SetAttr("touch_pair",ValueSet(val,cf))
	
class KSAttention(KnowledgeSource):
	def __init__(self):
		KnowledgeSource.__init__(self)
		self.attvList = []
		self.atttList = []
		self.maxAttvSo = None
		self.maxAtttSo = None
		
	def Step(self,model):
		#現在見えるものに対してAttention付加
		eyePos = model.eyePos
		model.attvList = []
		for lso in model.LSolids():
			lv,cf = lso.GetAttr("last_view")
			if lv == model.stepCount:
				pos,cfp = lso.GetAttr("position")
				velo,cfv = lso.GetAttr("velocity")
				if pos and velo:
					att_velo = velo.norm() * 0.1
					att_distance = 1 / (1 + ( pos - eyePos ).norm()) * 0.2
					attv = att_velo + att_distance
					lso.SetAttr("BUV_distance",ValueSet(att_distance,1.0))
					lso.SetAttr("BUV_velocity",ValueSet(att_velo,1.0))
					lso.SetAttr("BUV", ValueSet(attv,1.0))
					model.attvList += ((lso,attv),)# 0 -> lso : 1 -> attv 
			else:
				lso.DelAttr("BUV_distance")
				lso.DelAttr("BUV_velocity")
				lso.DelAttr("BUV")
				
		
		# 視覚注意リスト作成 & MaxAttention
		model.attvList = sorted( model.attvList , key= lambda i : i[1] , reverse=True )
		model.maxAttvSo = model.attvList[0][0] if len(model.attvList) > 0 else None
		
		# 現在触っているものに対してAttention付加
		model.atttList = []
		for lso in model.LSolids():
			lt,cf = lso.GetAttr("last_touch")
			if lt == model.stepCount:
				att_pressure,cf = lso.GetAttr("pressure")
				if not att_pressure:
					continue
			
				attt = att_pressure * 4.0
				lso.SetAttr("BUT_pressure",ValueSet(att_pressure,1.0))
				lso.SetAttr("BUT",ValueSet(attt,1.0))
				model.atttList += ((lso,attt),)# 0 -> lso : 1 -> attt
				
		# 触ってない場合はAttention = 0 にする(?)
			else:
				lso.DelAttr("BUT_pressure")
				lso.DelAttr("BUT")
				
			
		# 触覚注意リスト作成 & MaxAttention
		model.atttList = sorted( model.atttList , key= lambda i : i[1] , reverse=True )
		model.maxAtttSo = model.atttList[0][0] if len(model.atttList) > 0 else None
		
		# 視覚、触覚、あわせた注意の大きさ
		KV = 1.0
		KT = 0.1
		vision = KV * model.maxAttvSo.GetAttr("BUV").value if model.maxAttvSo else 0
		touch  = KT * model.maxAtttSo.GetAttr("BUT").value if model.maxAtttSo else 0
		if vision == touch == 0:
			model.maxAttSo = None
		else:
			if touch >= vision:
				model.maxAttSo = model.maxAtttSo
			else:
				model.maxAttSo = model.maxAttvSo

# AttrDBから情報を追加
class KSAttribute(KnowledgeSource):
	def __init__(self,cr,attrDB):
		KnowledgeSource.__init__(self)
		self.attrDB = attrDB
		self.cr = cr
	
	def Step(self,model):
		for lso in model.LSolids():
			try:
				self.attrDB.Interpret(lso,self.cr)
			except Exception as e:
				DSTR += str(e)+"\n"

# 推測
class KSPredict(KnowledgeSource):
	def Step(self,model):
		for lso in model.LSolids():
			lv,cfv = lso.GetAttr("last_view")
			lt,cft = lso.GetAttr("last_touch")
			# 見ている or 触ってる
			if lt == model.stepCount or lv == model.stepCount:
				memSo,cf = lso.GetAttr("memSo")
				if not memSo:
					memSo = model.GetInternalPHScene().CreateSolid()
					lso.SetAttr("memSo",ValueSet(memSo,1.0))
					#初回のみ同期
					# 形の同期
					cnt = 0
					for shape in lso.phSolid.GetShapes():
						memSo.AddShape(shape)
						memSo.SetShapePose(cnt, lso.phSolid.GetShapePose(cnt))
						cnt += 1
					#名前
					memSo.SetName( lso.phSolid.GetName() )
				#同期
				memSo.SetPose(lso.phSolid.GetPose())
				memSo.SetVelocity(lso.phSolid.GetVelocity())
				memSo.SetAngularVelocity( lso.phSolid.GetAngularVelocity())
				memSo.SetMass(lso.phSolid.GetMass())
				memSo.SetInertia(lso.phSolid.GetInertia())
				memSo.SetDynamical(lso.phSolid.IsDynamical())
				memSo.SetFramePosition(lso.phSolid.GetFramePosition())
			# 見ても触っても無い
			else:
				# 推測を元にlsoのattrを書き換える
				memSo,cf = lso.GetAttr("memSo")
				if memSo:
					val,cf = lso.GetAttr("position")
					val = memSo.GetFramePosition()
					lso.SetAttr("position",ValueSet(val,cf))
					
					
					val,cf = lso.GetAttr("velocity")
					val = memSo.GetVelocity()
					lso.SetAttr("velocity",ValueSet(val,cf))
					
class Model:
	def __init__(self,name):
		self.name = name
		self.__solids = {}
		self.__attrs = {}
		self.kses = [] #KnowledgeSource
	
	def __getitem__(self,key):
		try:
			ph = None
			if type(key) == Spr.PHSolid:
				ph = key
			elif type(key) == LSolid:
				ph = key.phSolid
			elif type(key) == str:
				return self.__attrs[key]
			
			return self.__solids[ph]
		except KeyError:
			return None
			
	
	def __setitem__(self,key,value):
		ph = None
		if type(key) == Spr.PHSolid:
			ph = key
		elif type(key) == LSolid:
			ph = key.phSolid
		elif type(key) == str:
			self.__attrs[key] = value
		else:
			raise KeyError("KeyはLSolidまたはPHSolidまたはstr限定です")
		
		if ph:
			self.__solids[ph] = value
		else:
			self.__attrs[key] = value
		
	def __contains__(self,key):
		ph = None
		if type(key) == Spr.PHSolid:
			ph = key
		elif type(key) == LSolid:
			ph = key.phSolid
		
		if ph and ph in self.__solids:
			return True
		elif key in self.__attrs:
			return True
		return False
	
	def LSolids(self):
		return self.__solids.values()
		
	def PHSolids(self):
		return self.__solids.keys()
	
	# Register Knowledge Source
	def RegisterKS(self,ks):
		self.kses.append(ks)
	
	def Step(self):
		for ks in self.kses:
			ks.Step(self)
	
	def Clear(self):
		self.__solids = {}
		
class CreatureModel(Model):
	def __init__(self,name,cr,phSdk):
		Model.__init__(self,name)
		self.cr = cr
		self.phSdk = phSdk
		self.phScene = phSdk.CreateScene()
	
	def GetInternalPHScene(self):
		return self.phScene
	
	def Clear(self):
		self.phSdk.DelChildObject(self.phScene)
		self.phScene = self.phSdk.CreateScene()
		Model.Clear(self)

