# -*- coding: utf-8 -*-
import copy

from .Utility import *

# task
#  PlanningBlackBoardのようなものをつくる。現在はdict型をそのまま使っているため
#  current["hoge"]としたとき、KeyErrorが出てしまうことがあるので、これをなくしたい
#  具体的にはEmptyObjectをかえすようにしたらいい？
#  あとGetSatisfactionRatio,GetCombinationをPlanningBBのメンバ関数にしてしまうといいかも

# plans同士のコンビネーションを取得する
def GetCombination(lhs,rhs):
	if len(rhs) == 0: return copy.copy(lhs)
	elif len(lhs) == 0: return copy.copy(rhs)
	
	ret = []
	for planl in lhs:
		for planr in rhs:
			ret.append( planl + planr )
	return ret

# ----------------------------------------------------
def GetSatisfactionRatio(current,goal):
	all = len(goal)
	satisifies = 0
	
	#goalに必要なものなしの場合,真(1)をかえす
	if all == 0:
		return 1
	
	for key,val in goal.items():
		#辞書に関数が入っていた場合
		if callable(val):
			#関数を条件関数(env を受け取り True or False を返す)としてあつかう 
			if val(current):
				#print(key,": 一致")
				satisifies += 1
			else:
				#print(key,": 不一致")
				continue
		#関数でない場合
		else:
			#キーが存在する場合
			if key in current:
				#値が違う場合、満たしていない
				if not current[key] == val:
					#print(key,": 不一致")
					continue
				#値が一致する場合、条件クリア数を増やす
				else:
					satisifies += 1
					#print(key,": 一致")
		
	return satisifies / all

class Planner:
	def __init__(self,cr):
		self.cr = cr
	
	def BuildPlan(self,goal):
		actions = []
		for lso in self.cr.models["current"].LSolids():
			afordance = lso.GetAttr("afordance").value
			if afordance:
				for act in lso.GetAttr("afordance").value:
					actions.append( (lso,act) )
		
		plans = self.BuildPlanR(goal,actions)
		
		# プランニング失敗(None)を含むプランを消す
		ret = []
		for plan in plans:
			
			if plan == None or None in plan:
				pass
			else:
				ret.append(plan)
		
		return ret
		
	
	def BuildPlanR(self,goal,actions,t=0):
		ratio = 0
		oldRatio = 0
		after = {} # 行動後の黒板
		actlist = []
		dprint("-- -- -- -- -- ")
		
		#goalを満たす行動リスト(「行動Aと行動Bをする必要がある」といったもの)を埋める
		for owner,act in actions:
			current = {owner.phSolid:owner.Clone()}
			effect = {}
			
			#今までにプランニングした行動のエフェクト
			effect.update( after )
			
			#対象行動のエフェクト
			effect.update( act.Effect(current) )
			
			#もし行動をしたらgoalを満たすかどうか
			#本来は評価関数でコストの評価などをするべきだが
			#とりあえずは最初に見つかったものから当てはめていく
			ratio = GetSatisfactionRatio(effect,goal)
			
			#もし対象行動をした場合、goal達成率が上がるかどうか
			if ratio > oldRatio:
				actlist.append((act,current))
				after.update( effect )
				oldRatio = ratio
			
			dprint(">",t,oldRatio,"->",ratio,act)
			dprint(">",t,"effect",effect)
			dprint(">",t,"after",after)
			dprint("")
		
		dprint("")
		dprint(">",t,"goal",goal)
		dprint(">",t,"actlist",actlist)
		
		#埋められた行動リストが条件をすべて満たしているか
		plans = []
		ratio = GetSatisfactionRatio(after,goal)
		if ratio >= 1:
			#リストを登録された順番で再帰的にプランニングしていく（これも評価すべき？）
			# A1 <= A3 , A2 <= None , A3 <= A4,A5
			# [A1,A2] -> [ [A3,A1], A2 ] -> [ [[A5,A4,A3],A1], A2 ]
			# 出てきたプランを左から順にやるととりあえず行動できるはず
			# 候補が複数出てくる(subplanがリストで帰ってくる)場合はすべての組み合わせを考える
			for act,current in actlist:
				subgoal = act.Precond()
				#サブゴールをもう満たしている場合は再帰打ち切り(展開しない)
				if GetSatisfactionRatio(current,subgoal) >= 1.0:
					dprint(">",t,"打ち切りT",act)
					plans = GetCombination(plans,[[(act,current)]])
				#満たしていない場合は再帰的にプランを展開して、現在のアクションの前に挿入
				else:
					dprint(">",t,"  再帰   ",act)
					dprint(">",t,"subgoal =",subgoal)
					dprint(">",t,"current =",current)
					dprint(">",t,"subgoal ratio = ",GetSatisfactionRatio(current,subgoal))
					subplans = self.BuildPlanR(subgoal,actions,t+1)
					composited = GetCombination(subplans,[[(act,current)]])
					dprint(">",t,composited)
					
					plans = GetCombination(plans, composited)
			return plans
		#すべて満たせない場合はプランニング不可能
		else:
			dprint(">",t,"打ち切りF(ratio=",ratio,")")
			return [[None]]

