# -*- coding: utf-8 -*-

import bpy
from mathutils import Vector,Quaternion

import spr_op.physics
from spr_op.physics import *

import spr_op.properties #propertyの定義
from spr_op.properties import register_properties

import time #Stepが過剰に呼ばれる問題の対処のため->L50


#PhysicsController-------------------------------------------------
physics = PhysicsController() 
print("PhysicsController physics: Created")



#Operator------------------------------------------
class SprOperator(bpy.types.Operator):
    '''Operator for Springhead2'''
    bl_idname = "spr.operator"
    bl_label = "SprOperator"
    _timer = None
    time_step = 1 / bpy.context.scene.game_settings.frequency
    
    def modal(self, context, event):
        
        #debug
        #print(event.type)
        
        #ESCイベントは使わない
        #if event.type == 'ESC':
            #print("spr.operator: stoped (ESC)")
            #return self.cancel(context)
        
        scene = bpy.context.scene
        
        if event.type == 'TIMER':#メインループ
            
            #info初期化
            physics.info["main"] = ""#step,syncの前にカラにする
            
            #time取得、計算
            time_step = self._timer.time_step #設定time step
            time_delta = self._timer.time_delta #Blender内部計測の微小時間
            
            ##Stepが過剰に呼ばれる現象の対策--------------------
            time_now = time.time()#self._timer.time_delta、self._timer.time_durationは当てにならないのでシステムtimeを呼ぶ
            actual_time_delta = time_now - physics.prev_time
            if actual_time_delta < time_delta:
                delta_of_time_delta = time_delta - actual_time_delta
                
                #debug
                #print("time_delta="+str(time_delta))
                #print("actual_time_delta="+str(actual_time_delta))
                #print("delta_of_time_delta="+str(delta_of_time_delta))
                
                if(delta_of_time_delta > time_step*0.1):#stepが過剰に呼ばれている場合 (設定time_step*0.1と比較)
                    #print("step is called redundantly")
                    return {'PASS_THROUGH'}
                    
            physics.prev_time = time_now
            
            
            
            #time step の更新---------------------------------------------
            if(self.time_step!=1 / scene.game_settings.frequency):#time stepが変わった時
                self.time_step = 1 / scene.game_settings.frequency
                #debug
                #print(self.time_step)
                
                context.window_manager.event_timer_remove(self._timer)
                self._timer = context.window_manager.event_timer_add(self.time_step, context.window)#60fps以上は変わらないので注意
            
            
            #同期---------------------------------------------------------
            physics.sync()
            
            #Springhead2 Step
            substep = scene.spr_sub_step#sub step この回数の物理ステップを回す
            physics_time_step = self.time_step/substep
            physics.set_scene_props(physics_time_step)
            
            if(scene.spr_step_enabled == 1):
                
                #Step Speed Factor ・・・現状挙動に影響が出てしまう模様・・・要デバッグ
                step_num = 0
                
                if scene.spr_step_speed_factor < 1 and scene.spr_step_speed_factor!=0:#Slow
                    slow_factor = int(1/scene.spr_step_speed_factor)
                    physics.slow_step_counter += 1
                    if physics.slow_step_counter >= slow_factor:
                        step_num = 1
                        physics.slow_step_counter = 0
                else:#Normal, Fast
                    step_num = int(scene.spr_step_speed_factor)
                
                #step_numの数だけStep
                i = 0
                while i < step_num:
                    
                    #Step
                    ##substep分だけstepを呼ぶ
                    j = 0
                    while j < substep:
                        physics.step()
                        
                        #run scrpt
                        if bpy.context.scene.spr_run_scripts_enabled:
                            physics.run_scripts(j, substep)
                        
                        j += 1
                    
                    ##Step KeyFrame (Record KeyFrameのため)
                    if(scene.spr_record_keyframe == 1):
                        
                        if(scene.spr_step_keyframe == 1):
                            if(scene.spr_use_unique_keyframe == 1):#独自のキーフレーム番号を用いる場合
                                scene.spr_unique_keyframe += 1
                                #step counter(unique keyframeを使う時しかうまくいかない)
                                physics.record_keyframe_step_counter += 1
                                if(physics.record_keyframe_step_counter >= scene.spr_record_every_n_steps):
                                    physics.record_keyframe_step_counter = 0    #次のStepでkeyframeを保存する
                            
                            else:#カレントキーフレームを用いる場合
                                scene.frame_current += 1
                                physics.record_keyframe_step_counter = 0    #次のStepでkeyframeを保存する
                    
                    #while文
                    i += 1
            
            
            #info-------------------------------------------------------
            ##平均FPSの計算
            physics.info_fps_list.append(1/time_delta)
            if(len(physics.info_fps_list) > 9):
                sum = 0
                for fps in physics.info_fps_list:
                    sum += fps
                physics.info_fps_ave = sum / len(physics.info_fps_list)
                del physics.info_fps_list[:]
            
            ##記録
            physics.info["main"] += "Target Time Step: " + str(time_step) + "\n"
            physics.info["main"] += "Actual Time Delta: " + str(actual_time_delta) + "\n"
            physics.info["main"] += "Actual FPS: " + str(physics.info_fps_ave) + "\n"
            physics.info["main"] += "Phisics Actual FPS: " + str(physics.info_fps_ave*substep) + "\n"
            physics.info["main"] += "Phisics Time Step: " + str(physics_time_step) + "\n"
            
            if not physics.phScene is None:
                physics.info["main"] += "Physics Step Count: " + str(physics.phScene.GetCount()) + "\n"
            
            
            #vizualize-------------------------
            physics.visualizer.update()
            
            
        return {'PASS_THROUGH'}

    def execute(self, context):#timer登録
        context.window_manager.modal_handler_add(self)
        self._timer = context.window_manager.event_timer_add(self.time_step, context.window)
        return {'RUNNING_MODAL'}

    def cancel(self, context):
        context.window_manager.event_timer_remove(self._timer)
        return {'CANCELLED'}


#ループ実行開始------------------------------------------------------------------------
def activate():
    bpy.ops.spr.operator()
#    print("spr.operator: started (ESC to stop)")
    print("spr.operator: started")


#UI--------------------------------------------------------------------------------------
##Object panel-------------------------------
class SprObjectPanel(bpy.types.Panel):
    bl_label = "Spr Object Panel"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "object"

    def draw(self, context):
        layout = self.layout

        obj = context.object
        mat = obj.material_slots.id_data.active_material
        
        #Real-Time Editing Marker
        icon_r = "NONE"
        if(obj.spr_solid_realtime_editing==1):
            icon_r = "FILE_TICK"
        
        #1段目：パラメータ適用・リアルタイムモード
        split = layout.split(align=True)
        row = split.row()
        row.prop(obj,"spr_solid_realtime_editing", text="Real-Time Editing", icon = "FILE_TICK")
        if(obj.spr_solid_realtime_editing==0):
            row.operator("spr.solid_parameters_apply",text = "Apply Parameters")
        
        #2段目：メイン
        split = layout.split()
        #左側：type,mass,radius,center
        col = split.column()
        #for solid
        col.label(text="Solid:")
        #col.prop(obj.game,"physics_type")
        #col.prop(obj.game, "collision_bounds_type", text="Bounds")
        col.prop(obj,"spr_physics_type", text="Physics Type", icon = icon_r)
        col.prop(obj,"spr_shape_type" ,text="Shape Type")
        if(obj.spr_shape_type=="Convex"):
            col.prop(obj,"spr_compound_child",text="Compound Child")
        if(obj.spr_compound_child == 0):
            row = col.row()
            row.alignment = "LEFT"
            row.label(text ="Mass:", icon = icon_r)
            row.prop(obj.game,"mass",text="")
            
            sub = col.column(align=True)
            sub.label(text = "Center:", icon = icon_r)
            sub.prop(obj,"spr_center_x",text="x", icon = icon_r)
            sub.prop(obj,"spr_center_y",text="y", icon = icon_r)
            sub.prop(obj,"spr_center_z",text="z", icon = icon_r)
            col.prop(obj,"spr_inertia_manual_setting_enabled", text="Inertia Manual Setting", icon = icon_r)
            if(obj.spr_inertia_manual_setting_enabled==1):
                sub = col.column(align=True)
                sub.prop(obj,"spr_inertia_x",text="x")
                sub.prop(obj,"spr_inertia_y",text="y")
                sub.prop(obj,"spr_inertia_z",text="z")
                
        #右側：material
        col = split.column()
        #for material
        col.label(text="Material:")
        if(mat!=None):
            row = col.row()
            row.alignment = "LEFT"
            row.label(text ="Elasticity:", icon = icon_r)
            row.prop(mat.physics,"elasticity", text="")
            
            row = col.row()
            row.alignment = "LEFT"
            row.label(text ="Friction:", icon = icon_r)
            row.prop(mat.physics,"friction", text="")
            
            row = col.row()
            row.alignment = "LEFT"
            row.label(text ="Static Friction:", icon = icon_r)
            row.prop(mat,"spr_friction_static",text="")
            
        ##icon TEST
        #icons = ['NONE', 'QUESTION', 'ERROR', 'CANCEL', 'TRIA_RIGHT', 'TRIA_DOWN', 'TRIA_LEFT', 'TRIA_UP', 'ARROW_LEFTRIGHT', 'PLUS', 'DISCLOSURE_TRI_DOWN', 'DISCLOSURE_TRI_RIGHT', 'RADIOBUT_OFF', 'RADIOBUT_ON', 'MENU_PANEL', 'BLENDER', 'DOT', 'X', 'GO_LEFT', 'PLUG', 'UI', 'NODE', 'NODE_SEL', 'FULLSCREEN', 'SPLITSCREEN', 'RIGHTARROW_THIN', 'BORDERMOVE', 'VIEWZOOM', 'ZOOMIN', 'ZOOMOUT', 'PANEL_CLOSE', 'COPY_ID', 'EYEDROPPER', 'LINK_AREA', 'AUTO', 'CHECKBOX_DEHLT', 'CHECKBOX_HLT', 'UNLOCKED', 'LOCKED', 'UNPINNED', 'PINNED', 'SCREEN_BACK', 'RIGHTARROW', 'DOWNARROW_HLT', 'DOTSUP', 'DOTSDOWN', 'LINK', 'INLINK', 'PLUGIN', 'HELP', 'GHOST_ENABLED', 'COLOR', 'LINKED', 'UNLINKED', 'HAND', 'ZOOM_ALL', 'ZOOM_SELECTED', 'ZOOM_PREVIOUS', 'ZOOM_IN', 'ZOOM_OUT', 'RENDER_REGION', 'BORDER_RECT', 'BORDER_LASSO', 'FREEZE', 'STYLUS_PRESSURE', 'GHOST_DISABLED', 'NEW', 'FILE_TICK', 'QUIT', 'URL', 'RECOVER_LAST', 'FULLSCREEN_ENTER', 'FULLSCREEN_EXIT', 'BLANK1', 'LAMP', 'MATERIAL', 'TEXTURE', 'ANIM', 'WORLD', 'SCENE', 'EDIT', 'GAME', 'RADIO', 'SCRIPT', 'PARTICLES', 'PHYSICS', 'SPEAKER', 'TEXTURE_SHADED', 'VIEW3D', 'IPO', 'OOPS', 'BUTS', 'FILESEL', 'IMAGE_COL', 'INFO', 'SEQUENCE', 'TEXT', 'IMASEL', 'SOUND', 'ACTION', 'NLA', 'SCRIPTWIN', 'TIME', 'NODETREE', 'LOGIC', 'CONSOLE', 'PREFERENCES', 'CLIP', 'ASSET_MANAGER', 'OBJECT_DATAMODE', 'EDITMODE_HLT', 'FACESEL_HLT', 'VPAINT_HLT', 'TPAINT_HLT', 'WPAINT_HLT', 'SCULPTMODE_HLT', 'POSE_HLT', 'PARTICLEMODE', 'LIGHTPAINT', 'SCENE_DATA', 'RENDERLAYERS', 'WORLD_DATA', 'OBJECT_DATA', 'MESH_DATA', 'CURVE_DATA', 'META_DATA', 'LATTICE_DATA', 'LAMP_DATA', 'MATERIAL_DATA', 'TEXTURE_DATA', 'ANIM_DATA', 'CAMERA_DATA', 'PARTICLE_DATA', 'LIBRARY_DATA_DIRECT', 'GROUP', 'ARMATURE_DATA', 'POSE_DATA', 'BONE_DATA', 'CONSTRAINT', 'SHAPEKEY_DATA', 'CONSTRAINT_BONE', 'PACKAGE', 'UGLYPACKAGE', 'BRUSH_DATA', 'IMAGE_DATA', 'FILE', 'FCURVE', 'FONT_DATA', 'RENDER_RESULT', 'SURFACE_DATA', 'EMPTY_DATA', 'SETTINGS', 'RENDER_ANIMATION', 'RENDER_STILL', 'BOIDS', 'STRANDS', 'LIBRARY_DATA_INDIRECT', 'GREASEPENCIL', 'GROUP_BONE', 'GROUP_VERTEX', 'GROUP_VCOL', 'GROUP_UVS', 'RNA', 'RNA_ADD', 'OUTLINER_OB_EMPTY', 'OUTLINER_OB_MESH', 'OUTLINER_OB_CURVE', 'OUTLINER_OB_LATTICE', 'OUTLINER_OB_META', 'OUTLINER_OB_LAMP', 'OUTLINER_OB_CAMERA', 'OUTLINER_OB_ARMATURE', 'OUTLINER_OB_FONT', 'OUTLINER_OB_SURFACE', 'OUTLINER_OB_SPEAKER', 'RESTRICT_VIEW_OFF', 'RESTRICT_VIEW_ON', 'RESTRICT_SELECT_OFF', 'RESTRICT_SELECT_ON', 'RESTRICT_RENDER_OFF', 'RESTRICT_RENDER_ON', 'OUTLINER_DATA_EMPTY', 'OUTLINER_DATA_MESH', 'OUTLINER_DATA_CURVE', 'OUTLINER_DATA_LATTICE', 'OUTLINER_DATA_META', 'OUTLINER_DATA_LAMP', 'OUTLINER_DATA_CAMERA', 'OUTLINER_DATA_ARMATURE', 'OUTLINER_DATA_FONT', 'OUTLINER_DATA_SURFACE', 'OUTLINER_DATA_SPEAKER', 'OUTLINER_DATA_POSE', 'MESH_PLANE', 'MESH_CUBE', 'MESH_CIRCLE', 'MESH_UVSPHERE', 'MESH_ICOSPHERE', 'MESH_GRID', 'MESH_MONKEY', 'MESH_CYLINDER', 'MESH_TORUS', 'MESH_CONE', 'LAMP_POINT', 'LAMP_SUN', 'LAMP_SPOT', 'LAMP_HEMI', 'LAMP_AREA', 'META_EMPTY', 'META_PLANE', 'META_CUBE', 'META_BALL', 'META_ELLIPSOID', 'META_CAPSULE', 'SURFACE_NCURVE', 'SURFACE_NCIRCLE', 'SURFACE_NSURFACE', 'SURFACE_NCYLINDER', 'SURFACE_NSPHERE', 'SURFACE_NTORUS', 'CURVE_BEZCURVE', 'CURVE_BEZCIRCLE', 'CURVE_NCURVE', 'CURVE_NCIRCLE', 'CURVE_PATH', 'FORCE_FORCE', 'FORCE_WIND', 'FORCE_VORTEX', 'FORCE_MAGNETIC', 'FORCE_HARMONIC', 'FORCE_CHARGE', 'FORCE_LENNARDJONES', 'FORCE_TEXTURE', 'FORCE_CURVE', 'FORCE_BOID', 'FORCE_TURBULENCE', 'FORCE_DRAG', 'MODIFIER', 'MOD_WAVE', 'MOD_BUILD', 'MOD_DECIM', 'MOD_MIRROR', 'MOD_SOFT', 'MOD_SUBSURF', 'HOOK', 'MOD_PHYSICS', 'MOD_PARTICLES', 'MOD_BOOLEAN', 'MOD_EDGESPLIT', 'MOD_ARRAY', 'MOD_UVPROJECT', 'MOD_DISPLACE', 'MOD_CURVE', 'MOD_LATTICE', 'CONSTRAINT_DATA', 'MOD_ARMATURE', 'MOD_SHRINKWRAP', 'MOD_CAST', 'MOD_MESHDEFORM', 'MOD_BEVEL', 'MOD_SMOOTH', 'MOD_SIMPLEDEFORM', 'MOD_MASK', 'MOD_CLOTH', 'MOD_EXPLODE', 'MOD_FLUIDSIM', 'MOD_MULTIRES', 'MOD_SMOKE', 'MOD_SOLIDIFY', 'MOD_SCREW', 'MOD_VERTEX_WEIGHT', 'MOD_DYNAMICPAINT', 'MOD_REMESH', 'REC', 'PLAY', 'FF', 'REW', 'PAUSE', 'PREV_KEYFRAME', 'NEXT_KEYFRAME', 'PLAY_AUDIO', 'PLAY_REVERSE', 'PREVIEW_RANGE', 'PMARKER_ACT', 'PMARKER_SEL', 'PMARKER', 'MARKER_HLT', 'MARKER', 'SPACE2', 'SPACE3', 'KEYINGSET', 'KEY_DEHLT', 'KEY_HLT', 'MUTE_IPO_OFF', 'MUTE_IPO_ON', 'VISIBLE_IPO_OFF', 'VISIBLE_IPO_ON', 'DRIVER', 'SOLO_OFF', 'SOLO_ON', 'FRAME_PREV', 'FRAME_NEXT', 'VERTEXSEL', 'EDGESEL', 'FACESEL', 'ROTATE', 'CURSOR', 'ROTATECOLLECTION', 'ROTATECENTER', 'ROTACTIVE', 'ALIGN', 'SMOOTHCURVE', 'SPHERECURVE', 'ROOTCURVE', 'SHARPCURVE', 'LINCURVE', 'NOCURVE', 'RNDCURVE', 'PROP_OFF', 'PROP_ON', 'PROP_CON', 'PARTICLE_POINT', 'PARTICLE_TIP', 'PARTICLE_PATH', 'MAN_TRANS', 'MAN_ROT', 'MAN_SCALE', 'MANIPUL', 'SNAP_OFF', 'SNAP_ON', 'SNAP_NORMAL', 'SNAP_INCREMENT', 'SNAP_VERTEX', 'SNAP_EDGE', 'SNAP_FACE', 'SNAP_VOLUME', 'STICKY_UVS_LOC', 'STICKY_UVS_DISABLE', 'STICKY_UVS_VERT', 'CLIPUV_DEHLT', 'CLIPUV_HLT', 'SNAP_PEEL_OBJECT', 'GRID', 'PASTEDOWN', 'COPYDOWN', 'PASTEFLIPUP', 'PASTEFLIPDOWN', 'SNAP_SURFACE', 'RETOPO', 'UV_VERTEXSEL', 'UV_EDGESEL', 'UV_FACESEL', 'UV_ISLANDSEL', 'UV_SYNC_SELECT', 'BBOX', 'WIRE', 'SOLID', 'SMOOTH', 'POTATO', 'ORTHO', 'LOCKVIEW_OFF', 'LOCKVIEW_ON', 'AXIS_SIDE', 'AXIS_FRONT', 'AXIS_TOP', 'NDOF_DOM', 'NDOF_TURN', 'NDOF_FLY', 'NDOF_TRANS', 'LAYER_USED', 'LAYER_ACTIVE', 'SORTALPHA', 'SORTBYEXT', 'SORTTIME', 'SORTSIZE', 'LONGDISPLAY', 'SHORTDISPLAY', 'GHOST', 'IMGDISPLAY', 'BOOKMARKS', 'FONTPREVIEW', 'FILTER', 'NEWFOLDER', 'FILE_PARENT', 'FILE_REFRESH', 'FILE_FOLDER', 'FILE_BLANK', 'FILE_BLEND', 'FILE_IMAGE', 'FILE_MOVIE', 'FILE_SCRIPT', 'FILE_SOUND', 'FILE_FONT', 'BACK', 'FORWARD', 'DISK_DRIVE', 'MATPLANE', 'MATSPHERE', 'MATCUBE', 'MONKEY', 'HAIR', 'ALIASED', 'ANTIALIASED', 'MAT_SPHERE_SKY', 'WORDWRAP_OFF', 'WORDWRAP_ON', 'SYNTAX_OFF', 'SYNTAX_ON', 'LINENUMBERS_OFF', 'LINENUMBERS_ON', 'SCRIPTPLUGINS', 'SEQ_SEQUENCER', 'SEQ_PREVIEW', 'SEQ_LUMA_WAVEFORM', 'SEQ_CHROMA_SCOPE', 'SEQ_HISTOGRAM', 'SEQ_SPLITVIEW', 'IMAGE_RGB', 'IMAGE_RGB_ALPHA', 'IMAGE_ALPHA', 'IMAGE_ZDEPTH', 'IMAGEFILE', 'BRUSH_ADD', 'BRUSH_BLOB', 'BRUSH_BLUR', 'BRUSH_CLAY', 'BRUSH_CLONE', 'BRUSH_CREASE', 'BRUSH_DARKEN', 'BRUSH_FILL', 'BRUSH_FLATTEN', 'BRUSH_GRAB', 'BRUSH_INFLATE', 'BRUSH_LAYER', 'BRUSH_LIGHTEN', 'BRUSH_MIX', 'BRUSH_MULTIPLY', 'BRUSH_NUDGE', 'BRUSH_PINCH', 'BRUSH_SCRAPE', 'BRUSH_SCULPT_DRAW', 'BRUSH_SMEAR', 'BRUSH_SMOOTH', 'BRUSH_SNAKE_HOOK', 'BRUSH_SOFTEN', 'BRUSH_SUBTRACT', 'BRUSH_TEXDRAW', 'BRUSH_THUMB', 'BRUSH_ROTATE', 'BRUSH_VERTEXDRAW', 'VIEW3D_VEC', 'EDIT_VEC', 'EDITMODE_DEHLT', 'EDITMODE_HLT', 'DISCLOSURE_TRI_RIGHT_VEC', 'DISCLOSURE_TRI_DOWN_VEC', 'MOVE_UP_VEC', 'MOVE_DOWN_VEC', 'X_VEC', 'SMALL_TRI_RIGHT_VEC']
        #for icon_ in icons:
        #    layout.label(text=icon_, icon = icon_)

###Joint
class SprJointPanel(bpy.types.Panel):
    bl_label = "Spr Joint Panel"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "object"
    
    
    
    def draw(self, context):
        layout = self.layout
        
        obj = context.object
        scene = context.scene
        
        #Real-Time Editing Marker
        icon_r = "NONE"
        if(obj.spr_joint_realtime_editing==1):
            icon_r = "FILE_TICK"
        
        #1段目：パラメータ適用・リアルタイムモード
        split = layout.split(align=True)
        row = split.row()
        row.prop(obj,"spr_joint_realtime_editing", text="Real-Time Editing", icon = "FILE_TICK")
        if(obj.spr_joint_realtime_editing==0):
            row.operator("spr.joint_parameters_apply",text = "Apply Parameters")
        
        #2段目：Additional Joint
        #Additional Joint----
        box = layout.box()
        row = box.row()
        row.prop(obj,"spr_additional_joint", text="Additional Joint")
        if(obj.spr_additional_joint==0):
            row.operator("spr.add_joint",text = "Add Joint")
        else:
            try:
                parent_name = obj.parent.name
            except:
                parent_name = "(None)"
            row.label(text="This is Additional Joint of " + parent_name)
            
        
        
        #3段目：パラメータ
        split = layout.split()
        #左側：type, limit
        col = split.column()
        col.prop(obj,"spr_joint_type", text="Joint Type")
        if(obj.spr_joint_type!="None" and obj.spr_compound_child == 0):#Jointを作る場合のみ以下を表示する
            
            col.prop(obj,"spr_joint_enabled", text="Enable",icon = icon_r)
            
            col.prop(obj,"spr_joint_target", text="Joint Target Name")
            
            if(obj.spr_joint_type=="Spring"):
                sub = col.column(align=True)
                sub.label(text ="Spring:", icon = icon_r)
                sub.prop(obj,"spr_spring_x", text="x")
                sub.prop(obj,"spr_spring_y", text="y")
                sub.prop(obj,"spr_spring_z", text="z")
                sub.prop(obj,"spr_spring_ori", text="ori")
                sub = col.column(align=True)
                sub.label(text ="Damper:", icon = icon_r)
                sub.prop(obj,"spr_damper_x", text="x")
                sub.prop(obj,"spr_damper_y", text="y")
                sub.prop(obj,"spr_damper_z", text="z")
                sub.prop(obj,"spr_damper_ori", text="ori")
            else:
                row = col.row()
                row.alignment = "LEFT"
                row.label(text ="Spring:", icon = icon_r)
                row.prop(obj,"spr_spring", text="")
                
                row = col.row()
                row.alignment = "LEFT"
                row.label(text ="Damper:", icon = icon_r)
                row.prop(obj,"spr_damper", text="")
            
            col.prop(obj,"spr_joint_collision", text="Joint Collision")
            
            
            col.separator()
            
            #limit
            col.label(text="Limit", icon = icon_r)
            col.prop(obj,"spr_limit_spring", text="Limit Spring")
            col.prop(obj,"spr_limit_damper", text="Limit Damper")
            if(obj.spr_joint_type=="Ball"):
                #swing
                sub = col.column(align=True)
                sub.label(text="Limit Swing:", icon = icon_r)
                
                row = sub.row()
                row.alignment = "LEFT"
                row.prop(obj,"spr_limit_swing_max",text="Max")
                row.operator("spr.limit_inf_swing_max",text = "inf")
                
                row = sub.row()
                row.alignment = "LEFT"
                row.prop(obj,"spr_limit_swing_min",text="min")
                row.operator("spr.limit_inf_swing_min",text = "inf")
                
                #twist
                sub = col.column(align=True)
                sub.label(text="Limit Twist:", icon = icon_r)
                
                row = sub.row()
                row.alignment = "LEFT"
                row.prop(obj,"spr_limit_twist_max",text="Max")
                row.operator("spr.limit_inf_twist_max",text = "inf")

                row = sub.row()
                row.alignment = "LEFT"
                row.prop(obj,"spr_limit_twist_min",text="min")
                row.operator("spr.limit_inf_twist_min",text = "inf")
                
            if(obj.spr_joint_type=="Hinge" or obj.spr_joint_type=="Slider"):
                sub = col.column(align=True)
                if(obj.spr_joint_type=="Hinge"):
                    sub.label(text="Range:", icon = icon_r)
                else:
                    sub.label(text="Range:", icon = icon_r)
                
                row = sub.row()
                row.alignment = "LEFT"
                row.prop(obj,"spr_limit_range_max",text="Max")
                row.operator("spr.limit_inf_range_max",text = "inf")
                
                row = sub.row()
                row.alignment = "LEFT"
                row.prop(obj,"spr_limit_range_min",text="min")
                row.operator("spr.limit_inf_range_min",text = "inf")
                
            
            col.separator()
            
            #Deformation(Elastic-Plastic)
            col.label(text="Deformation(Elastic-Plastic)")
            row = col.row()
            row.alignment = "LEFT"
            row.label(text ="Yield Stress:", icon = icon_r)
            row.prop(obj,"spr_yield_stress",text="")
            
            if(obj.spr_joint_type=="Ball"):
                sub = col.column(align=True)
                sub.label(text="Second Damper:", icon = icon_r)
                sub.prop(obj,"spr_second_damper_x",text="x")
                sub.prop(obj,"spr_second_damper_y",text="y")
                sub.prop(obj,"spr_second_damper_z",text="z")
            
            elif(obj.spr_joint_type=="Hinge" or obj.spr_joint_type=="Slider"):
                row = col.row()
                row.alignment = "LEFT"
                row.label(text ="Second Damper:", icon = icon_r)
                row.prop(obj,"spr_second_damper",text="")
            
            row = col.row()
            row.alignment = "LEFT"
            row.label(text ="Hardness Rate:", icon = icon_r)
            row.prop(obj,"spr_hardness_rate",text="")
            
            
            #右側：plug socket
            col = split.column()
            #Plug----
            col.label(text="Plug:")
            if(obj.spr_additional_joint==0):#追加のJointではない場合
                if(obj.spr_joint_plug_handle==0):
                    col.operator("spr.plug_handle_on",text = "Plug Handle Mode")
                    
                    sub = col.column(align=True)#表示のみ。編集はplugオブジェクト(Empty)で
                    sub.label(text="Plug Pos:")
                    sub.label(text="x: "+str(obj.spr_joint_plug_pos_x))
                    sub.label(text="y: "+str(obj.spr_joint_plug_pos_y))
                    sub.label(text="z: "+str(obj.spr_joint_plug_pos_z))
                    
                    sub = col.column(align=True)
                    sub.label(text="Plug Ori:")
                    sub.label(text="w: "+str(obj.spr_joint_plug_ori_w))
                    sub.label(text="x: "+str(obj.spr_joint_plug_ori_x))
                    sub.label(text="y: "+str(obj.spr_joint_plug_ori_y))
                    sub.label(text="z: "+str(obj.spr_joint_plug_ori_z))
                    
                    sub = col.column(align=True)
                    sub.label(text="Plug Ori Offset(deg):")
                    sub.prop(obj,"spr_joint_plug_offset_x",text="x")
                    sub.prop(obj,"spr_joint_plug_offset_y",text="y")
                    sub.prop(obj,"spr_joint_plug_offset_z",text="z")
                    
                elif(obj.spr_joint_plug_handle==1):
                    col.operator("spr.plug_handle_apply",text = "Apply Plug Handle")
                    
                    obj_plug_handle = None
                    obj_plug_handle_working = 0
                    try:
                        obj_plug_handle = bpy.data.objects["p_"+obj.name]
                        obj_plug_handle_working = 1
                    except:
                        col.label(text="Error! Please Push [Apply Plug Handle]")
                    
                    if(obj_plug_handle_working == 1):
                        col.label(text="Plug Handle: ON")
            else:#追加のJointの場合
                col.label(text="Use Pose of This Object as Plug-Pose")
                loc = obj.location
                rot = obj.rotation_quaternion #(w,x,y,z)
                
                sub = col.column(align=True)#表示
                sub.label(text="Plug Pos:")
                sub.label(text="x: "+str(loc.x))
                sub.label(text="y: "+str(loc.y))
                sub.label(text="z: "+str(loc.z))
                
                sub = col.column(align=True)
                sub.label(text="Plug Ori:")
                sub.label(text="w: "+str(rot.w))
                sub.label(text="x: "+str(rot.x))
                sub.label(text="y: "+str(rot.y))
                sub.label(text="z: "+str(rot.z))
                
                sub = col.column(align=True)
                sub.label(text="Plug Ori Offset(deg):")
                sub.prop(obj,"spr_joint_plug_offset_x",text="x")
                sub.prop(obj,"spr_joint_plug_offset_y",text="y")
                sub.prop(obj,"spr_joint_plug_offset_z",text="z")
            
            
            col.separator()
            
            #Socket----
            col.prop(obj,"spr_joint_socket_enabled",text="Set Socket")
            if obj.spr_joint_socket_enabled == 1:
                if(obj.spr_joint_socket_handle==0):
                    col.operator("spr.socket_handle_on",text = "Socket Handle Mode")
                    
                    sub = col.column(align=True)#表示のみ。編集はsocketオブジェクト(Empty)で
                    sub.label(text="Socket Pos:")
                    sub.label(text="x: "+str(obj.spr_joint_socket_pos_x))
                    sub.label(text="y: "+str(obj.spr_joint_socket_pos_y))
                    sub.label(text="z: "+str(obj.spr_joint_socket_pos_z))
                    
                    sub = col.column(align=True)
                    sub.label(text="Socket Ori:")
                    sub.label(text="w: "+str(obj.spr_joint_socket_ori_w))
                    sub.label(text="x: "+str(obj.spr_joint_socket_ori_x))
                    sub.label(text="y: "+str(obj.spr_joint_socket_ori_y))
                    sub.label(text="z: "+str(obj.spr_joint_socket_ori_z))
                    
                elif(obj.spr_joint_socket_handle==1):
                    box.operator("spr.socket_handle_apply",text = "Apply Socket Handle")
                    
                    obj_socket_handle = None
                    obj_socket_handle_working = 0
                    try:
                        obj_socket_handle = bpy.data.objects["s_"+obj.name]
                        obj_socket_handle_working = 1
                    except:
                        box.label(text="Error! Please Push [Apply Socket Handle]")
                    
                    if(obj_socket_handle_working == 1):
                        box.label(text="Socket Handle: ON")
        
        #4段目：制御
        box = layout.box()
        split = box.split()
        split.label(text="Joint Control")
        split = box.split()
        if(obj.spr_joint_type=="Hinge" or obj.spr_joint_type=="Slider"):#1自由度関節Hinge, Slider
            #左列
            col = split.column()
            col.label("Target Position",icon = icon_r)
            col.prop(obj,"spr_joint_target_position", text="")
            col.label("Target Velocity",icon = icon_r)
            col.prop(obj,"spr_joint_target_velocity", text="")
            
            #右列
            col = split.column()
            col.label("Offset Force",icon = icon_r)
            col.prop(obj,"spr_joint_offset_force", text="")
            col.label("Max Force",icon = icon_r)
            col.prop(obj,"spr_joint_max_force", text="")
            
        elif(obj.spr_joint_type=="Ball"):#Ball Joint
            #左列
            col = split.column()
            sub=col.column(align=True)
            sub.label("Target Position(deg)",icon = icon_r)
            sub.prop(obj,"spr_joint_target_position_x", text="x")
            sub.prop(obj,"spr_joint_target_position_y", text="y")
            sub.prop(obj,"spr_joint_target_position_z", text="z")
            
            sub=col.column(align=True)
            sub.label("Target Velocity(deg)",icon = icon_r)
            sub.prop(obj,"spr_joint_target_velocity_x", text="x")
            sub.prop(obj,"spr_joint_target_velocity_y", text="y")
            sub.prop(obj,"spr_joint_target_velocity_z", text="z")
            
            #右列
            col = split.column()
            sub=col.column(align=True)
            sub.label("Offset Force",icon = icon_r)
            sub.prop(obj,"spr_joint_offset_force_x", text="x")
            sub.prop(obj,"spr_joint_offset_force_y", text="y")
            sub.prop(obj,"spr_joint_offset_force_z", text="z")
            
            col.label("Max Force",icon = icon_r)
            col.prop(obj,"spr_joint_max_force", text="")
            
        else:
            col = split.column()
            col.label("This Joint Type is Not Supported")
        
        
###IK
class SprIKPanel(bpy.types.Panel):
    bl_label = "Spr IK Panel"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "object"

    def draw(self, context):
    
        layout = self.layout
        obj = context.object
        scene = context.scene
        
        #Real-Time Editing Marker
        icon_r = "NONE"
        if(obj.spr_ik_realtime_editing==1):
            icon_r = "FILE_TICK"
        
        #1段目：パラメータ適用・リアルタイムモード
        split = layout.split(align=True)
        row = split.row()
        row.prop(obj,"spr_ik_realtime_editing", text="Real-Time Editing", icon = "FILE_TICK")
        if(obj.spr_ik_realtime_editing==0):
            row.operator("spr.ik_parameters_apply",text = "Apply Parameters")
        
        #2段目：パラメータ
        split = layout.split()
        col = split.column()
        col.prop(obj,"spr_ik_enabled",text="IK Enabled")
        if(obj.spr_ik_enabled == 1 and obj.spr_compound_child == 0):
            #左側：IK Actuator
            col.label(text="IK Actuator")
            row = col.row()
            row.alignment = "LEFT"
            row.label(text ="IK Bias:", icon = icon_r)
            row.prop(obj,"spr_ik_bias", text="")
            
            #右側：IK EndEffector
            col = split.column()
            col.prop(obj,"spr_ik_end_effector", text="IK EndEffector")
            if(obj.spr_ik_end_effector == 1):
                col.label("Target Local Position", icon = icon_r)
                sub = col.column(align=True)
                sub.prop(obj,"spr_ik_target_local_pos_x", text="x")
                sub.prop(obj,"spr_ik_target_local_pos_y", text="y")
                sub.prop(obj,"spr_ik_target_local_pos_z", text="z")
            
            
        
        if obj.spr_ik_end_effector == 1:
            box = layout.box()
            split = box.split()
            split.label(text="IK Control")
            #3段目：制御 Target指定方法選択
            split = box.split()
            row = split.row()
            row.prop(obj,"spr_ik_pos_control_enabled", text="Position Control", icon = icon_r)
            row.prop(obj,"spr_ik_ori_control_enabled", text="Orientation Control", icon = icon_r)
            split = box.split()
            split.prop(obj,"spr_ik_target_object_enabled", text="Enable Target Object", icon = icon_r)
            
            #4段目：制御 Target設定
            split = box.split()
            if obj.spr_ik_target_object_enabled == 1:#目標にobjectの姿勢を用いる場合
                split.prop(obj,"spr_ik_target_object_name", text="Target Object Name", icon = icon_r)
            else:#目標を数値で設定する場合
                col = split.column()
                sub = col.column(align=True)
                sub.label(text="Target Position:", icon = icon_r)
                sub.prop(obj,"spr_ik_target_pos_x",text="x")
                sub.prop(obj,"spr_ik_target_pos_y",text="y")
                sub.prop(obj,"spr_ik_target_pos_z",text="z")
                col = split.column()
                sub = col.column(align=True)
                sub.label(text="Target Orientation(deg):", icon = icon_r)
                sub.prop(obj,"spr_ik_target_ori_x",text="x")
                sub.prop(obj,"spr_ik_target_ori_y",text="y")
                sub.prop(obj,"spr_ik_target_ori_z",text="z")


###Solid Control
class SprSolidControlPanel(bpy.types.Panel):
    bl_label = "Spr Solid Control Panel"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "object"
    
    def draw(self, context):
    
        layout = self.layout
        obj = context.object
        scene = context.scene
        
        #1段目：Solid Info
        layout.prop(obj,"spr_solid_info_enabled", text="Solid Info")
        if(obj.spr_solid_info_enabled==1):
            split = layout.split()
            if obj.name in physics.solid_controllers:
                if not (physics.solid_controllers[obj.name].phSolid is None):
                    #値取得
                    solid = physics.solid_controllers[obj.name].phSolid
                    v_x = str(solid.GetVelocity().x)
                    v_y = str(solid.GetVelocity().y)
                    v_z = str(solid.GetVelocity().z)
                    ang_v_x = str(solid.GetAngularVelocity().x)
                    ang_v_y = str(solid.GetAngularVelocity().y)
                    ang_v_z = str(solid.GetAngularVelocity().z)
                    f_x = str(solid.GetForce().x)
                    f_y = str(solid.GetForce().y)
                    f_z = str(solid.GetForce().z)
                    t_x = str(solid.GetTorque().x)
                    t_y = str(solid.GetTorque().y)
                    t_z = str(solid.GetTorque().z)
                    
                    #左列
                    col = split.column()
                    col.label("[Velocity]")
                    col.label("x: " + v_x)
                    col.label("y: " + v_y)
                    col.label("z: " + v_z)
                    
                    col.label("[Angular V]")
                    col.label("x: " + ang_v_x)
                    col.label("y: " + ang_v_y)
                    col.label("z: " + ang_v_z)
                    
                    #右列
                    col = split.column()
                    col.label("[Force]")
                    col.label("x: " + f_x)
                    col.label("y: " + f_y)
                    col.label("z: " + f_z)
                    
                    col.label("[Torque]")
                    col.label("x: " + t_x)
                    col.label("y: " + t_y)
                    col.label("z: " + t_z)
                    
            else:
                col = split.column()
                col.label("Solid is Invalid")
        
        #2段目：パラメータ適用
        split = layout.split()
        
        #左列
        col = split.column()
        
        sub = col.column(align=True)
        sub.label(text="Velocity")
        row = sub.row()
        row.prop(obj,"spr_solid_velocity_realtime_control_enabled", text="realtime")
        if obj.spr_solid_velocity_realtime_control_enabled == 0:
            row.operator("spr.solid_control_velocity_apply",text = "Apply")
        sub.prop(obj,"spr_solid_velocity_x", text="x")
        sub.prop(obj,"spr_solid_velocity_y", text="y")
        sub.prop(obj,"spr_solid_velocity_z", text="z")
        sub.prop(obj,"spr_solid_velocity_local", text="local")
        
        
        sub = col.column(align=True)
        sub.label(text="Angular V")
        row = sub.row()
        row.prop(obj,"spr_solid_ang_velocity_realtime_control_enabled", text="realtime")
        if obj.spr_solid_ang_velocity_realtime_control_enabled == 0:
            row.operator("spr.solid_control_angular_velocity_apply",text = "Apply")
        sub.prop(obj,"spr_solid_angular_velocity_x", text="x")
        sub.prop(obj,"spr_solid_angular_velocity_y", text="y")
        sub.prop(obj,"spr_solid_angular_velocity_z", text="z")
        sub.prop(obj,"spr_solid_ang_velocity_local", text="local")
        
        #右列
        col = split.column()
        
        sub = col.column(align=True)
        sub.label(text="Add Force")
        row = sub.row()
        row.prop(obj,"spr_solid_force_realtime_control_enabled", text="realtime")
        if obj.spr_solid_force_realtime_control_enabled == 0:
            row.operator("spr.solid_control_force_apply",text = "Apply")
        sub.prop(obj,"spr_solid_force_x", text="x")
        sub.prop(obj,"spr_solid_force_y", text="y")
        sub.prop(obj,"spr_solid_force_z", text="z")
        sub.prop(obj,"spr_solid_set_force_point", text="Set Force Point")
        if(obj.spr_solid_set_force_point==1):
            sub.prop(obj,"spr_solid_force_point_x", text="x")
            sub.prop(obj,"spr_solid_force_point_y", text="y")
            sub.prop(obj,"spr_solid_force_point_z", text="z")
        else:
            sub.prop(obj,"spr_solid_force_local", text="local")
        
        sub = col.column(align=True)
        sub.label(text="Add Torque")
        row = sub.row()
        row.prop(obj,"spr_solid_torque_realtime_control_enabled", text="realtime")
        if obj.spr_solid_torque_realtime_control_enabled == 0:
            row.operator("spr.solid_control_torque_apply",text = "Apply")
        sub.prop(obj,"spr_solid_torque_x", text="x")
        sub.prop(obj,"spr_solid_torque_y", text="y")
        sub.prop(obj,"spr_solid_torque_z", text="z")
        sub.prop(obj,"spr_solid_torque_local", text="local")
        
        
        
###for Creature
class SprCreaturePanel(bpy.types.Panel):
    bl_label = "Spr Creature Panel"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "object"

    def draw(self, context):
        layout = self.layout

        obj = context.object
        
        
        layout.prop(obj,"spr_creature_body", text="Creature Body")
        if(obj.spr_creature_body==1 and obj.spr_compound_child == 0):
            layout.prop(obj,"spr_creature_id", text="Creature ID")
            layout.prop(obj,"spr_CRIKSolid_label", text="CRIKSolid Label")
            layout.prop(obj,"spr_CRIKJoint_label", text="CRIKJoint Label")
            layout.prop(obj,"spr_creature_core", text="Creature Core")
            
            if(obj.spr_creature_core>0):
                row = layout.row()
                row.prop(obj,"spr_creature_realtime_editing", text="Real-Time Editing", icon = "FILE_TICK")
                if(obj.spr_creature_realtime_editing==0):
                    row.operator("spr.cr_parameters_apply",text = "Apply Parameters")
                
                row = layout.row()
                row.prop(obj,"spr_creature_spring_ratio", text="Spring Ratio")
                row.prop(obj,"spr_creature_damper_ratio", text="Damper Ratio")
                
                row = layout.row()
                row.prop(obj,"spr_creature_view_angle_alt", text="View Angle Alt")
                row.prop(obj,"spr_creature_view_angle_az", text="View Angle Az")
                
                split = layout.split()
                col = split.column()
                col.label(text="Rules")
                col = split.column()
                col.operator("spr.cr_rules_apply",text = "Apply Rules")
                
                split = layout.split()
                split.label(text="Rule Parameters (Real-Time)", icon = "FILE_TICK")
                split = layout.split()
                col = split.column()
                col.prop(obj,"spr_creature_rule_parameter_0", text="p0")
                col.prop(obj,"spr_creature_rule_parameter_1", text="p1")
                col.prop(obj,"spr_creature_rule_parameter_2", text="p2")
                col = split.column()
                col.prop(obj,"spr_creature_rule_parameter_3", text="p3")
                col.prop(obj,"spr_creature_rule_parameter_4", text="p4")
                col.prop(obj,"spr_creature_rule_parameter_5", text="p5")
                
                split = layout.split()
                split.prop(obj,"spr_creature_visualize_enabled", text="Visualizer")
                split = layout.split()
                col = split.column()
                col.operator("spr.cr_visualize_display",text = "Visualize Display")
                col = split.column()
                col.operator("spr.cr_visualize_hide",text = "Visualize Hide")
                
                split = layout.split()
                split.label(text="Visualizer Scale")
                split = layout.split()
                col = split.column()
                col.prop(obj,"spr_visualizer_scale_base", text="Base Scale")
                col.prop(obj,"spr_visualizer_scale_max", text="Max Scale")
                col = split.column()
                col.prop(obj,"spr_visualizer_attv_scale_f", text="Attv Factor")
                col.prop(obj,"spr_visualizer_attt_scale_f", text="Attt Factor")
                

##Material panel-------------------------------
#Spr Object Panelでも設定できるがMaterialパネルにも表示する
class SprMaterialPanel(bpy.types.Panel):
    bl_label = "Spr Material Panel"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "material"

    def draw(self, context):
        layout = self.layout

        obj = context.object
        mat = obj.material_slots.id_data.active_material
        
        #Real-Time Editing Marker
        icon_r = "NONE"
        if(obj.spr_solid_realtime_editing==1):
            icon_r = "FILE_TICK"
        
        #1段目：パラメータ適用・リアルタイムモード
        split = layout.split()
        row = split.row()
        row.prop(obj,"spr_solid_realtime_editing", text="Real-Time Editing", icon = "FILE_TICK")
        if(obj.spr_solid_realtime_editing==0):
            row.operator("spr.solid_parameters_apply",text = "Apply Parameters")
        
        #2段目：メイン
        split = layout.split()
        col = split.column()
        #for material
        col.label(text="Material:")
        if(mat!=None):
            row = col.row()
            row.alignment = "LEFT"
            row.label(text ="Elasticity:", icon = icon_r)
            row.prop(mat.physics,"elasticity", text="")
            
            row = col.row()
            row.alignment = "LEFT"
            row.label(text ="Friction:", icon = icon_r)
            row.prop(mat.physics,"friction", text="")
            
            row = col.row()
            row.alignment = "LEFT"
            row.label(text ="Static Friction:", icon = icon_r)
            row.prop(mat,"spr_friction_static",text="")

##Scene Panel---------------------------------
class SprScenePanel(bpy.types.Panel):
    bl_label = "Spr Scene Panel"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "scene"

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        
        layout.prop(scene, "gravity")
        split = layout.split()
        col = split.column()
        col.prop(scene.game_settings, "frequency", text="FPS", slider=False)
        col = split.column()
        col.prop(scene, "spr_sub_step", text="SubSteps", slider=False)
        
        split = layout.split()
        split.label(text="LCP Num Iteration (default=15)")
        split.prop(scene, "spr_num_iteration", text="NumIteration", slider=False)
        
        
##UI Panel-------------------------------------
class SprOperatorPanel(bpy.types.Panel):
    bl_label = "Spr Operator Panel"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"

    def draw(self, context):
        layout = self.layout
        scene = context.scene
        
        col = layout.column(align=True)
        col.label(text="Physics Step:")
        col.prop(scene, "spr_step_enabled", text = "Enable")
        col.prop(scene, "spr_step_speed_factor", text = "Step Speed Factor")
        if scene.spr_step_enabled == 0:
            col.operator("spr.one_step",text = "One Step")
        
        #layout.separator()
        
        col = layout.column(align=True)
        col.label(text="Set Physics:")
        col.operator("spr.initialize",text = "Set Scene")
        col.operator("spr.add_object",text = "Set Objects")
        
        #layout.separator()
        
        col = layout.column(align=True)
        col.label(text = "Restore Scene:")
        col.operator("spr.restore",text = "Restore Scene", icon="RECOVER_LAST")
        
        #layout.separator()
        
        col = layout.column(align=True)
        col.label(text = "Options:")
        col.prop(scene, "spr_throw_enabled", text = "Enable Throw")
        col.prop(scene, "spr_ik_engine_enabled", text = "Enable IK Engine")
        col.prop(scene, "spr_ik_iteration_num", text = "IK Iteration(per step)")
        col.prop(scene, "spr_creature_enabled", text = "Enable Creature")
        
        #layout.separator()
        
        col = layout.column(align=True)
        col.label(text = "Run Scripts:")
        col.prop(scene, "spr_run_scripts_enabled", text = "Run")
        col.operator("spr.update_scripts",text = "Update Scripts")
        
        #layout.separator()
        
        col = layout.column(align=True)
        col.label(text = "Utilities:")
        col.operator("spr.re_register",text = "ReRegister")
        col.operator("spr.save_spr",text = "Save SPR File")
        
        #layout.separator()
        
        col = layout.column(align=True)
        col.label(text = "Keyframe", icon="IPO")
        col.prop(scene, "spr_record_keyframe", text = "Record KeyFrame(Loc, Rot)", icon="REC")
        col.prop(scene, "spr_step_keyframe", text = "Step Keyframe")
        if scene.spr_step_keyframe == 1:
            col.prop(scene, "spr_use_unique_keyframe", text = "Use Unique Keyframe")
            if scene.spr_use_unique_keyframe == 1:
                col.prop(scene, "spr_unique_keyframe", text = "Unique Keyframe")
                col.prop(scene, "spr_record_every_n_steps", text = "Record Every N Steps")



#UI buttons------------------------------------------------------------------------------------
##for UI Panel----
class OBJECT_OT_SprOneStep(bpy.types.Operator):
    bl_idname = "spr.one_step"
    bl_label = "[spr]Call One Step"
    bl_description = "[Spr]Call ONE physics step."
    
    def execute(self,context):
        physics.step()
        
        return{'FINISHED'}

class OBJECT_OT_SprInitialize(bpy.types.Operator):
    bl_idname = "spr.initialize"
    bl_label = "[spr]Initialize whole scene"
    bl_description = "[Spr]Initialize and set whole physics scene (phSdk, phScene, phSolid...etc.)"
    
    def execute(self,context):
        init()
        
        return{'FINISHED'}

class OBJECT_OT_SprActivate(bpy.types.Operator):
    bl_idname = "spr.activate"
    bl_label = "[spr]Activate Spr Operator"
    bl_description = "[Spr]"
    
    def execute(self,context):
        activate()
        
        return{'FINISHED'}

class OBJECT_OT_SprRestore(bpy.types.Operator):
    bl_idname = "spr.restore"
    bl_label = "[spr]Restore scene"
    bl_description = "[Spr]Restore pose of objects in the scene to initial state."
    
    def execute(self,context):
        physics.restore()
        
        return{'FINISHED'}
        
class OBJECT_OT_SprAddObject(bpy.types.Operator):
    bl_idname = "spr.add_object"
    bl_label = "[spr]set Objects"
    bl_description = "[Spr]Set selected objects to physics scene."
    
    def execute(self,context):
        physics.add_object()
        
        return{'FINISHED'}
        

class OBJECT_OT_UpdateScripts(bpy.types.Operator):
    bl_idname = "spr.update_scripts"
    bl_label = "[spr]Update Scripts"
    bl_description = "[Spr]Update user scripts."
    
    def execute(self,context):
        physics.update_scripts()
        
        return{'FINISHED'}
        

class OBJECT_OT_SprSaveSpr(bpy.types.Operator):
    bl_idname = "spr.save_spr"
    bl_label = "[spr]Save spr"
    bl_description = "[Spr]Save spr file."
    
    def execute(self,context):
        physics.save_spr()
        
        return{'FINISHED'}
        
class OBJECT_OT_SprReRegister(bpy.types.Operator):
    bl_idname = "spr.re_register"
    bl_label = "[spr]ReRegister and activate"
    bl_description = "[Spr]Push at ONCE after opening the other file."
    
    def execute(self,context):
        re_register()
        
        return{'FINISHED'}

##for Creature Panel----
class OBJECT_OT_SprCrVisualizeDisplay(bpy.types.Operator):
    bl_idname = "spr.cr_visualize_display"
    bl_label = "[spr]Creature Visualize Display"
    bl_description = "[Spr]"
    
    def execute(self,context):
        
        id = context.object.spr_creature_id
        
        physics.visualizer.cr_visualize_display_switch(id, False)
        
        return{'FINISHED'}

class OBJECT_OT_SprCrVisualizeHide(bpy.types.Operator):
    bl_idname = "spr.cr_visualize_hide"
    bl_label = "[spr]Creature Visualize Hide"
    bl_description = "[Spr]"
    
    def execute(self,context):
        
        id = context.object.spr_creature_id
        
        physics.visualizer.cr_visualize_display_switch(id, True)
        
        return{'FINISHED'}

class OBJECT_OT_SprCrRulesApply(bpy.types.Operator):
    bl_idname = "spr.cr_rules_apply"
    bl_label = "[spr]Creature Apply Rules"
    bl_description = "[Spr]"
    
    def execute(self,context):
        
        id = context.object.spr_creature_id
        
        if id in physics.creature_controllers:
            physics.creature_controllers[id].update_rules()
        
        return{'FINISHED'}

class OBJECT_OT_SprCrParametersApply(bpy.types.Operator):
    bl_idname = "spr.cr_parameters_apply"
    bl_label = "[spr]Creature Apply Parameters"
    bl_description = "[Spr]"
    
    def execute(self,context):
        
        id = context.object.spr_creature_id
        
        if id in physics.creature_controllers:
            physics.creature_controllers[id].update_parameters()
        
        return{'FINISHED'}

##for Solid Control Panel----
class OBJECT_OT_SprSolidControlVelocityApply(bpy.types.Operator):
    bl_idname = "spr.solid_control_velocity_apply"
    bl_label = "[spr]Solid Control Velocity Apply"
    bl_description = "[Spr]"
    
    def execute(self,context):
        obj = context.object
        
        if obj.name in physics.solid_controllers:
            physics.solid_controllers[obj.name].solid_control_apply(0)
        
        return{'FINISHED'}
        
class OBJECT_OT_SprSolidControlAngularVelocityApply(bpy.types.Operator):
    bl_idname = "spr.solid_control_angular_velocity_apply"
    bl_label = "[spr]Solid Control Angular Velocity Apply"
    bl_description = "[Spr]"
    
    def execute(self,context):
        obj = context.object
        
        if obj.name in physics.solid_controllers:
            physics.solid_controllers[obj.name].solid_control_apply(1)
        
        return{'FINISHED'}
        
class OBJECT_OT_SprSolidControlForceApply(bpy.types.Operator):
    bl_idname = "spr.solid_control_force_apply"
    bl_label = "[spr]Solid Control Force Apply"
    bl_description = "[Spr]"
    
    def execute(self,context):
        obj = context.object
        
        if obj.name in physics.solid_controllers:
            physics.solid_controllers[obj.name].solid_control_apply(2)
        
        return{'FINISHED'}

class OBJECT_OT_SprSolidControlTorqueApply(bpy.types.Operator):
    bl_idname = "spr.solid_control_torque_apply"
    bl_label = "[spr]Solid Control Torque Apply"
    bl_description = "[Spr]"
    
    def execute(self,context):
        obj = context.object
        
        if obj.name in physics.solid_controllers:
            physics.solid_controllers[obj.name].solid_control_apply(3)
        
        return{'FINISHED'}



##for Solid Panel
class OBJECT_OT_SprSolidParametersApply(bpy.types.Operator):
    bl_idname = "spr.solid_parameters_apply"
    bl_label = "[spr]Apply Solid Parameters"
    bl_description = "[Spr]"
    
    def execute(self,context):
        obj = context.object
        
        if obj.name in physics.solid_controllers:
            so_con = physics.solid_controllers[obj.name]
            so_con.set_solid_parameters()
            so_con.set_shape_parameters()
        
        return{'FINISHED'}



##for IK Panel----
class OBJECT_OT_SprIKParametersApply(bpy.types.Operator):
    bl_idname = "spr.ik_parameters_apply"
    bl_label = "[spr]Apply IK Parameters"
    bl_description = "[Spr]"
    
    def execute(self,context):
        obj = context.object
        
        if obj.name in physics.solid_controllers:
            so_con = physics.solid_controllers[obj.name]
            so_con.set_ika_parameters()
            so_con.set_ikeff_parameters()
        
        return{'FINISHED'}

##fot Joint Panel----
###Realtime
class OBJECT_OT_SprJointParametersApply(bpy.types.Operator):
    bl_idname = "spr.joint_parameters_apply"
    bl_label = "[spr]Apply Joint Parameters"
    bl_description = "[Spr]"
    
    def execute(self,context):
        obj = context.object
        jo_obj = obj
        if obj.spr_additional_joint == 1:
            obj = obj.parent
            if obj is None:
                return{'FINISHED'}
        
        if obj.name in physics.solid_controllers:
            so_con = physics.solid_controllers[obj.name]
            
            if jo_obj.name in so_con.joint_controllers:
                jo_con = so_con.joint_controllers[jo_obj.name]
                
                jo_con.set_joint_parameters()
                
        
        return{'FINISHED'}

###plug handle
#####memo: panelのclassではオブジェクトの変更は基本的に許されない -> 外部やボタンのclassならOK
class OBJECT_OT_SprPlugHandleOn(bpy.types.Operator): 
    bl_idname = "spr.plug_handle_on"
    bl_label = "[spr]joint plug handle ON"
    bl_description = "[Spr]"
    
    def execute(self,context):
        
        obj = context.object
        scene = context.scene
        
        obj_plug_handle = None
        
        try:#既存なら既存のものを使う
            obj_plug_handle = bpy.data.objects["p_"+obj.name]
            #Crear Parent Inverse
            bpy.ops.object.select_all(action='DESELECT')
            obj_plug_handle.hide_select = False
            obj_plug_handle.select = True
            bpy.ops.object.parent_clear(type="CLEAR_INVERSE")
            
        except:#なければ作成
            obj_plug_handle = bpy.data.objects.new("p_"+obj.name, None)
            scene.objects.link(obj_plug_handle)
            obj_plug_handle.empty_draw_type = "ARROWS"
            obj_plug_handle.parent = obj
            bpy.ops.object.select_all(action='DESELECT')
            obj_plug_handle.select = True
            
        #位置姿勢をデータに合わせる
        obj_plug_handle.location = (obj.spr_joint_plug_pos_x, obj.spr_joint_plug_pos_y, obj.spr_joint_plug_pos_z)
        obj_plug_handle.rotation_mode = "QUATERNION"
        obj_plug_handle.rotation_quaternion = Quaternion((obj.spr_joint_plug_ori_w, obj.spr_joint_plug_ori_x, obj.spr_joint_plug_ori_y, obj.spr_joint_plug_ori_z))
        
        #handle modeフラグ
        obj.spr_joint_plug_handle = 1
        
        return{'FINISHED'}

class OBJECT_OT_SprPlugHandleApply(bpy.types.Operator): 
    bl_idname = "spr.plug_handle_apply"
    bl_label = "[spr]joint plug handle Apply"
    bl_description = "[Spr]"
    
    def execute(self,context):
        
        obj = context.object
        scene = context.scene
        
        obj_plug_handle = None
        
        try:#既存のオブジェクトを取得
            obj_plug_handle = bpy.data.objects["p_"+obj.name]
        except:
            #handle modeフラグ
            obj.spr_joint_plug_handle = 0
            return{'FINISHED'}
        
        #Apply
        loc = obj_plug_handle.location
        rot = obj_plug_handle.rotation_quaternion#(w,x,y,z)
        
        obj.spr_joint_plug_pos_x = loc.x
        obj.spr_joint_plug_pos_y = loc.y
        obj.spr_joint_plug_pos_z = loc.z
        
        obj.spr_joint_plug_ori_w = rot.w
        obj.spr_joint_plug_ori_x = rot.x
        obj.spr_joint_plug_ori_y = rot.y
        obj.spr_joint_plug_ori_z = rot.z
        
        #選択不能にする(削除はしない)
        obj_plug_handle.hide_select = True
        
        #handle modeフラグ
        obj.spr_joint_plug_handle = 0
        
        return{'FINISHED'}

###socket handle
####memo: panelのclassではオブジェクトの変更は基本的に許されない -> 外部やボタンのclassならOK
class OBJECT_OT_SprSocketHandleOn(bpy.types.Operator): 
    bl_idname = "spr.socket_handle_on"
    bl_label = "[spr]joint socket handle ON"
    bl_description = "[Spr]"
    
    def execute(self,context):
        
        obj = context.object
        scene = context.scene
        
        obj_socket_handle = None
        
        try:#既存なら既存のものを使う
            obj_socket_handle = bpy.data.objects["s_"+obj.name]
            #Crear Parent Inverse
            bpy.ops.object.select_all(action='DESELECT')
            obj_socket_handle.hide_select = False
            obj_socket_handle.select = True
            bpy.ops.object.parent_clear(type="CLEAR_INVERSE")
            
        except:#なければ作成
            obj_socket_handle = bpy.data.objects.new("s_"+obj.name, None)
            scene.objects.link(obj_socket_handle)
            obj_socket_handle.empty_draw_type = "ARROWS"
            bpy.ops.object.select_all(action='DESELECT')
            obj_socket_handle.select = True

        
        #joint親オブジェクトにparent
        obj_patent = None
        try:#joint親オブジェクト取得
            obj_parent = bpy.data.objects[obj.spr_joint_target]
            obj_socket_handle.parent = obj_parent
        except:
            #print("Error: Joint Target Object is Missing")
            return{'FINISHED'}
        
        #位置姿勢をデータに合わせる
        obj_socket_handle.location = (obj.spr_joint_socket_pos_x, obj.spr_joint_socket_pos_y, obj.spr_joint_socket_pos_z)
        obj_socket_handle.rotation_mode = "QUATERNION"
        obj_socket_handle.rotation_quaternion = Quaternion((obj.spr_joint_socket_ori_w, obj.spr_joint_socket_ori_x, obj.spr_joint_socket_ori_y, obj.spr_joint_socket_ori_z))
        
        #handle modeフラグ
        obj.spr_joint_socket_handle = 1
        
        return{'FINISHED'}

class OBJECT_OT_SprSocketHandleApply(bpy.types.Operator): 
    bl_idname = "spr.socket_handle_apply"
    bl_label = "[spr]joint socket handle Apply"
    bl_description = "[Spr]"
    
    def execute(self,context):
        
        obj = context.object
        scene = context.scene
        
        obj_socket_handle = None
        
        try:#既存のオブジェクトを取得
            obj_socket_handle = bpy.data.objects["s_"+obj.name]
        except:
            #handle modeフラグ
            obj.spr_joint_socket_handle = 0
            return{'FINISHED'}
        
        #Apply
        loc = obj_socket_handle.location
        rot = obj_socket_handle.rotation_quaternion#(w,x,y,z)
        
        obj.spr_joint_socket_pos_x = loc.x
        obj.spr_joint_socket_pos_y = loc.y
        obj.spr_joint_socket_pos_z = loc.z
        
        obj.spr_joint_socket_ori_w = rot.w
        obj.spr_joint_socket_ori_x = rot.x
        obj.spr_joint_socket_ori_y = rot.y
        obj.spr_joint_socket_ori_z = rot.z
        
        #選択不能にする(削除はしない)
        obj_socket_handle.hide_select = True
        
        #handle modeフラグ
        obj.spr_joint_socket_handle = 0
        
        return{'FINISHED'}

###Limit Infinit (limit解除ボタン)
class OBJECT_OT_SprLimitInfRangeMax(bpy.types.Operator):
    bl_idname = "spr.limit_inf_range_max"
    bl_label = "[spr]Limit Inf Range Max"
    bl_description = "[Spr]"
    
    def execute(self,context):
        obj = context.object
        obj.spr_limit_range_max = 3e+39
        
        return{'FINISHED'}

class OBJECT_OT_SprLimitInfRangeMin(bpy.types.Operator):
    bl_idname = "spr.limit_inf_range_min"
    bl_label = "[spr]Limit Inf Range min"
    bl_description = "[Spr]"
    
    def execute(self,context):
        obj = context.object
        obj.spr_limit_range_min = 3e+39
        
        return{'FINISHED'}

class OBJECT_OT_SprLimitInfSwingMax(bpy.types.Operator):
    bl_idname = "spr.limit_inf_swing_max"
    bl_label = "[spr]Limit Inf Swing Max"
    bl_description = "[Spr]"
    
    def execute(self,context):
        obj = context.object
        obj.spr_limit_swing_max = 3e+39
        
        return{'FINISHED'}

class OBJECT_OT_SprLimitInfSwingMin(bpy.types.Operator):
    bl_idname = "spr.limit_inf_swing_min"
    bl_label = "[spr]Limit Inf Swing min"
    bl_description = "[Spr]"
    
    def execute(self,context):
        obj = context.object
        obj.spr_limit_swing_min = 3e+39
        
        return{'FINISHED'}

class OBJECT_OT_SprLimitInfTwistMax(bpy.types.Operator):
    bl_idname = "spr.limit_inf_twist_max"
    bl_label = "[spr]Limit Inf twist Max"
    bl_description = "[Spr]"
    
    def execute(self,context):
        obj = context.object
        obj.spr_limit_twist_max = 3e+39
        
        return{'FINISHED'}

class OBJECT_OT_SprLimitInfTwistMin(bpy.types.Operator):
    bl_idname = "spr.limit_inf_twist_min"
    bl_label = "[spr]Limit Inf twist min"
    bl_description = "[Spr]"
    
    def execute(self,context):
        obj = context.object
        obj.spr_limit_twist_min = 3e+39
        
        return{'FINISHED'}

class OBJECT_OT_SprAddJoint(bpy.types.Operator):
    bl_idname = "spr.add_joint"
    bl_label = "[spr]Add Joint"
    bl_description = "[Spr]Add Joint as Empty, Child of This Object"
    
    def execute(self,context):
        obj = context.object
        scene = context.scene
        
        obj_joint = None
        
        i = 0
        while 1:
            if "jo"+str(i)+obj.name in bpy.data.objects:
                i += 1
            else:
                obj_joint = bpy.data.objects.new("jo"+str(i)+obj.name, None)
                scene.objects.link(obj_joint)
                obj_joint.empty_draw_type = "ARROWS"
                obj_joint.parent = obj
                obj_joint.rotation_mode = "QUATERNION"
                
                obj_joint.spr_additional_joint = 1
                obj_joint.spr_joint_type = obj.spr_joint_type
                
                
                break
        
        return{'FINISHED'}


#register-------------------------------------------------------------------------------------------
def register():
    bpy.utils.register_module(__name__)

    register_properties()

    print("spr.operator,ui,properties: Registed")


def unregister():
    bpy.utils.unregister_module(__name__)

    print("spr.operator,ui: UnRegisted")


#at first(register)----------------------------------------------------------------------------------

register()

physics.initialize()#PHSdk,PHSceneだけ作っておく

activate()



#便利関数・コマンド----------------------------------------------------------------------------------

#def run():
#    physics.initialize()
#    physics.construct_scene()
#    activate()

#別ファイルロード後初期化(再登録)用
def re_register():
    unregister()
    
    register()
    physics.initialize()
    activate()


#実行中の初期化用
def init():
    physics.initialize()
    physics.construct_scene()
    
#koguma取得用
def get_koguma():
    return physics.creature_controllers[1].koguma

