# -*- coding: utf-8 -*-
#日本語UTF-8

import Spr
import time
from imp import reload

from .Creature.LSolid import LSolid
from .Creature.Model import Model
from .Creature.Database import *
from .Creature.Planner import *


def rad(deg):
    return deg / 180.0 * 3.1415926

#---------------
import bpy
from mathutils import Vector,Quaternion

#--------------PyKoguma creature.py 13 - 85 start--------------
def BuildWorldDB(db):
    #root 
    item = DatabaseItem()
    item.value = {}
    item.name ="root"
    db.items[item.name] = item
    db.root = item
    
    parent = item                #### root↓
    #root 
    item = DatabaseItem()
    item.value = {"shape":"plane"}
    item.name ="soFloor"
    db.items[item.name] = item
    db.root = item
    
    # Apple
    item = DatabaseItem()
    item.value = {"shape":"sphere"}
    item.parents += (parent,)
    parent.children += (item,)
    item.name = "soApple"
    db.items[item.name] = item
    
    parent = item                ######## Apple↓
    # RipeApple
    item = DatabaseItem()
    item.value = {"color":"red","taste":"sweet"}
    item.parents += (parent,)
    parent.children += (item,)
    item.name = "soAppleR"
    db.items[item.name] = item
    # UnripeApple
    item = DatabaseItem()
    item.value = {"color":"green","taste":"acid"}
    item.parents += (parent,)
    parent.children += (item,)
    item.name = "soAppleG"
    db.items[item.name] = item

def BuildAttrDB(db):
    #LoadFromFile(...)    #いずれ

    #root
    item = Attribute()
    item.value = "root-value"
    item.name = "root"
    db.items[item.name] = item
    db.root = item
    
    parent = item                ####root↓
    #food
    item = Attribute()
    item.value = {"eatable":50}
    item.afordance = ["eat","cook"]
    item.parents += (parent,)
    parent.children += (item,)
    item.name = "food"
    db.items[item.name] = item
    
    parent = item                ########food↓
    #fruit
    item = Attribute()
    item.value = {"eatable":80}
    item.afordance = ["liquidize"]
    item.value = {}
    item.parents += (parent,)
    parent.children += (item,)
    item.name = "fruit"
    db.items[item.name] = item

#--------------PyKoguma creature.py 13 - 85 end--------------

#==============CreatureController==========================================
class CreatureController():
    def __init__(self,ph_con,id):
        self.ph_con = ph_con
        self.id = id
        
        self.creature = None
        #self.koguma = None
        self.cr = None
        
        self.ready = False
        self.core_object = None
        
        self.soDebugOut = None
        
        self.body_solids = []
        
    def update_rules(self):
        #初期化
        self.cr.rules.clear()
        
        try:#ルールテキストが既存ならそちらを使う
            rule_text = bpy.data.texts["cr_" + str(self.id) +"_0_Basic"]#例：cr_0_0_Basic
        except:#無ければデフォルトの物を作る
            rule_text = bpy.data.texts.new("cr_" + str(self.id) +"_0_Basic")
            #以下デフォルトのルール（サンプル）
            rule = "#" + "cr_"+ str(self.id) + "_0_Basic" + "\n"
            rule += "#" + "This is Basic Rule for cr_" + str(self.id) + "\n"
            rule += "#" + "To Use Multi Rules, Create Rule-Name-List as \"cr_" + str(self.id) + "_rules\"" + "\n"
            rule += "#" + "------------------------------" + "\n"
            rule += "#" + "Rule Properties" + "\n"
            rule += "p = [1,1,1,1,1,1]" + "\n"
            rule += "import bpy" + "\n"
            rule += "if \"Base\" in bpy.data.objects:" + "\n"
            rule += "    p[0] = bpy.data.objects[\"Base\"].spr_creature_rule_parameter_0" + "\n"
            rule += "    p[1] = bpy.data.objects[\"Base\"].spr_creature_rule_parameter_1" + "\n"
            rule += "    p[2] = bpy.data.objects[\"Base\"].spr_creature_rule_parameter_2" + "\n"
            rule += "    p[3] = bpy.data.objects[\"Base\"].spr_creature_rule_parameter_3" + "\n"
            rule += "    p[4] = bpy.data.objects[\"Base\"].spr_creature_rule_parameter_4" + "\n"
            rule += "    p[5] = bpy.data.objects[\"Base\"].spr_creature_rule_parameter_5" + "\n"
            rule += "" + "\n"
            rule += "#" + "------------------------------" + "\n"
            rule += "currentModel = self.models[\"current\"]" + "\n"
            rule += "maxAttvSo = currentModel.maxAttvSo" + "\n"
            rule += "maxAtttSo = currentModel.maxAtttSo" + "\n"
            rule += "maxAttSo = currentModel.maxAttSo" + "\n"
            rule += "if maxAttSo:" + "\n"
            rule += "    self.HeadController.LookAt(maxAttSo.phSolid.GetFramePosition(),Spr.Vec3d(),1)" + "\n"
            rule += "    self.LeftHand.SetTargetPos(maxAttSo.phSolid.GetFramePosition())" + "\n"
            rule += "    self.RightHand.SetTargetPos(maxAttSo.phSolid.GetFramePosition())" + "\n"
            rule += "    self.LeftHand.Start()" + "\n"
            rule += "    self.RightHand.Start()" + "\n"
            
            rule_text.from_string(rule)
            
        self.cr.rules["cr_" + str(self.id) +"_0_Basic"] = rule_text.as_string()
        
        #複数Textに対応
        if("cr_" + str(self.id) +"_rules" in bpy.data.texts):#例：cr_0_rules
            for line in bpy.data.texts["cr_" + str(self.id) +"_rules"].lines:
                rule_name = line.body
                if(rule_name in bpy.data.texts):
                    self.cr.rules[rule_name] = bpy.data.texts[rule_name].as_string()
        
        
    def modify_spring_damper(self):
        spring_ratio = self.core_object.spr_creature_spring_ratio
        damper_ratio = self.core_object.spr_creature_damper_ratio
        
        body = self.creature.GetBody(0)
        for crikj in body.GetJoints():
            crikj_name = crikj.GetName()
            obj_name = crikj_name.lstrip("crikj")
            obj = self.ph_con.solid_controllers[obj_name].object
            #crikj.GetIKActuator().SetSpring(obj.spr_ik_spring * spring_ratio)#IK spring damper廃止
            #crikj.GetIKActuator().SetDamper(obj.spr_ik_damper * damper_ratio )
            crikj.GetPHJoint().SetSpring(obj.spr_spring * spring_ratio )
            crikj.GetPHJoint().SetDamper(obj.spr_damper * damper_ratio )
    
    def update_parameters(self):
        #gravity
        g = bpy.context.scene.gravity
        self.cr.models["current"].phScene.SetGravity(Spr.Vec3f(g.x,g.y,g.z))
        
        #spring damper ratio
        #cr1.ModifySpringDamper(0.5,2)
        #spring_ratio = self.core_object.spr_creature_spring_ratio
        #damper_ratio = self.core_object.spr_creature_damper_ratio
        #self.cr.ModifySpringDamper(spring_ratio, damper_ratio) #適用した回数だけ掛けちゃうからダメ
        self.modify_spring_damper()
        
        #view angle
        view_angle_alt = self.core_object.spr_creature_view_angle_alt
        view_angle_az = self.core_object.spr_creature_view_angle_az
        self.cr.VisualSensor.SetViewArea(Spr.Posed(0.710107,-0.710107,0,0, 0,0,0), rad(view_angle_az*2), rad(view_angle_alt*2))

    
    def create_creature(self):
        
        crSdk = self.ph_con.crSdk
        
        self.soDebugOut = self.ph_con.phScene.CreateSolid(Spr.PHSolidDesc())
        
        #CRCreature--------------
        crDesc = Spr.CRCreatureDesc()
        self.creature = crSdk.CreateCreature(Spr.CRCreature().GetIfInfoStatic(),crDesc)
        self.creature.SetName("creature"+str(self.id))
        self.creature.AddChildObject(self.ph_con.phScene)
        
        #CRBody-------------------
        crbDesc = Spr.CRBodyDesc()
        body = self.creature.CreateBody(Spr.CRBody().GetIfInfoStatic(),crbDesc)
        body.SetName("creature"+str(self.id)+"body")
        
        #Creature構成-----------------
        for name,so_con in self.ph_con.solid_controllers.items():
            if so_con.object.spr_creature_body == 1 and so_con.object.spr_creature_id == self.id:
                #debug
                #print(so_con.object.name)
                #phSolid = so_con.phSolid
                #ikeff = so_con.ikeff
                #phJoint = so_con.phJoint
                #ika = so_con.ika
                #if(not phSolid is None):
                #    print("solid OK")
                #if(not ikeff is None):
                #    print("ikeff OK")
                #if(not phJoint is None):
                #    print("joint OK")
                #if(not ika is None):
                #    print("ika OK")
                
                #core
                if(so_con.object.spr_creature_core==1):
                    self.core_object = so_con.object
                    
                #Solid
                criksDesc = Spr.CRIKSolidDesc()
                criksDesc.label = so_con.object.spr_CRIKSolid_label
                
                
                #if (not so_con.phSolid is None) and (not so_con.ikeff is None):
                if (not so_con.phSolid is None):#CRIKSolidにEndEffectorは必須ではない
                    crikSolid = body.CreateObject(Spr.CRIKSolid().GetIfInfoStatic(), criksDesc)
                    crikSolid.SetName("criks"+so_con.object.name)
                    crikSolid.AddChildObject(so_con.phSolid)
                    if (not so_con.ikeff is None):
                        crikSolid.AddChildObject(so_con.ikeff)
                    
                    self.body_solids.append(so_con.phSolid)
                    
                    #debug
                    #print("CRIKSolid Label: "+crikSolid.GetLabel())
                
                #Joint
                crikjDesc = Spr.CRIKJointDesc()
                crikjDesc.label = so_con.object.spr_CRIKJoint_label
                phJoint = None
                if not so_con.object is None:
                    if so_con.object.name in so_con.joint_controllers:
                        phJoint = so_con.joint_controllers[so_con.object.name].phJoint
                if (not phJoint is None) and (not so_con.ika is None):
                    crikJoint = body.CreateObject(Spr.CRIKJoint().GetIfInfoStatic(), crikjDesc)
                    crikJoint.SetName("crikj"+so_con.object.name)
                    crikJoint.AddChildObject(phJoint)
                    crikJoint.AddChildObject(so_con.ika)
                    
                    #debug
                    #print("CRIKJoint Label: "+crikJoint.GetLabel())
                
        #CRNeckController
        neckCtlDesc = Spr.CRNeckControllerDesc()
        #neckCtlDesc.lowerAttractiveness = 0.0
        #neckCtlDesc.upperAttractiveness = 0.5
        neckCtlDesc.labelHeadSolid = "Head"
        neckCtl = self.creature.CreateObject(Spr.CRNeckController().GetIfInfoStatic(), neckCtlDesc)
        neckCtl.SetName("neckCtl")
        
        
        #身体の相互接触を切る
        i = 0
        n = len(self.body_solids)
        print(n)
        while i < n:
            j = i + 1
            while j < n:
                self.ph_con.phScene.SetContactMode(self.body_solids[i], self.body_solids[j], 0) #ContactMode 0:MODE_NONE 1:MODE_PENALTY 2:MODE_LCP
                j += 1
            i += 1
        
        
        #準備判定
        self.ready = True
        try:
            body = self.creature.GetBody(0)
            print(body.GetName())
        except:
            print("Can Not Get Body! ID: "+str(self.id))
            return()
        
        if not (self.core_object is None):
            print("core: " +self.core_object.name )
        else:
            self.ready = False
            print("core is None!")
        
        crik = body.FindByLabel("Head")
        if not (crik is None):
            print("Label:Head " + crik.GetName())
        else:
            self.ready = False
            print("Head is None!")
        
        crik = body.FindByLabel("RightHand")
        if not (crik is None):
            print("Label:RightHand " + crik.GetName())
        else:
            self.ready = False
            print("RightHand is None!")
        
        crik = body.FindByLabel("LeftHand")
        if not (crik is None):
            print("Label:LeftHand " + crik.GetName())
        else:
            self.ready = False
            print("LeftHand is None!")
        
        crik = body.FindByLabel("Waist")
        if not (crik is None):
            print("Label:Waist " + crik.GetName())
        else:
            self.ready = False
            print("Waist is None!")
        
        crik = body.FindByLabel("Neck")
        if not (crik is None):
            print("Label:Neck " + crik.GetName())
        else:
            self.ready = False
            print("Neck is None!")
        
        crik = body.FindByLabel("RightElbow")
        if not (crik is None):
            print("Label:RightElbow " + crik.GetName())
        else:
            self.ready = False
            print("RightElbow is None!")
        
        crik = body.FindByLabel("LeftElbow")
        if not (crik is None):
            print("Label:LeftElbow " + crik.GetName())
        else:
            self.ready = False
            print("LeftElbow is None!")
        
        
        if self.ready == False :
            print("Create Creature Failed")
        else:
            
            #--------------PyKoguma creature.py 89 - 129 (modified) start--------------
            
            #given
            phSdk = self.ph_con.phSdk
            phScene = phSdk.GetScene(0)
            #crSdk = crSdk
            
            # -- -- -- -- -- データベースを構築
            worldDB = WorldDatabase("world")
            BuildWorldDB(worldDB)
            #print("**worldDB**")
            #worldDB.PrintDB()
            
            attrDB = AttrDatabase("attr")
            BuildAttrDB(attrDB)
            #print("**attrDB**")
            #attrDB.PrintDB()
            
            
            
            # -- -- -- -- -- -- 世界モデルを PHSolidから LSolidをつくり入れる -- -- -- -- 
            worldModel = Model("World model")
            print("Create LabeledSolids")
            for so in phScene.GetSolids():
                try:
                    lso = LSolid(so)
                    worldDB.Interpret(lso)
                    worldModel.solids[so] = lso
                except Exception as e:
                    print(e)
            
            
            
            # -- -- -- -- -- -- Creatureセットアップ -- -- --- -- --
            if crSdk.NCreatures() < 1:
                raise Exception("No Creature exists!!")
            crCreature = crSdk.GetCreatures()[0]
            
            from .Creature.Creature import Creature
            cr1 = Creature.CreateFromCRCreature( crCreature, phSdk, attrDB, worldModel, self.soDebugOut)
            
            
                        
            self.cr = cr1
            self.update_rules()#Creatureの行動ルールをセットアップ
            self.update_parameters()#パラメータをセットアップ
            
            #--------------PyKoguma creature.py 89 - 129 (modified) end--------------
            
            #--------------PyKoguma creature.py 230+α - start--------------
            
            phSceneMemory = cr1.models["current"].phScene
            phScene.GetIKEngine().Enable(True)
            cr1.start()
            
            #--------------PyKoguma creature.py 241 - end--------------
            
            self.cr = cr1
            


