# -*- coding: utf-8 -*-
#日本語UTF-8

import bpy

import Spr

import spr_op.bl_util as bl_util

import math
from math import sin, cos, tan, pi, radians

import copy

class Visualizer():
    def __init__(self,ph_con):
        self.ph_con = ph_con
        
        self.visualize_cr_view_angle = True
        self.visualize_cr_current_model = True
        
        #debug_out
        self.visualize_debug_out = True
        
        #view angle
        self.prev_view_angle_alt = -1  #Altitude 仰俯角・高度
        self.prev_view_angle_az = -1   #Azimuth 方位角
        
        
    def update(self):
        
        if self.visualize_cr_view_angle == True:
            self.update_view_angle()
        
        if self.visualize_cr_current_model == True:
            self.update_cr_current_model()
        
        if self.visualize_debug_out == True:
            self.update_debug_out()
    
    
    #creature current model
    def update_cr_current_model(self):
        
        context_obj = bpy.context.object
        
        
        for id,cr_con in self.ph_con.creature_controllers.items():
            if(cr_con.ready == True):
                
                #ステップ毎に邪魔にならないところに退避
                if "visualize_obj_"+str(id) in bpy.data.groups:
                    group = bpy.data.groups["visualize_obj_"+str(id)]
                    for name,obj in group.objects.items():
                        if not obj.name[0:9] == "spr_view_":
                            obj.location = (0,0,-10**16)
                
                #scale
                scale_base = cr_con.core_object.spr_visualizer_scale_base #default: 1.0
                scale_max = cr_con.core_object.spr_visualizer_scale_max #default: 5.0
                scale_attv_factor = cr_con.core_object.spr_visualizer_attv_scale_f #default: 3.0
                scale_attt_factor = cr_con.core_object.spr_visualizer_attt_scale_f #default: 0.0001
                
                
                info_cr_lso = ""
                info_cr_attv = ""
                info_cr_attt = ""
                info_cr_max = ""
                
                info = "== == == Creature Info == == ==\n"
                
                cr = cr_con.cr
                currentModel = copy.copy(cr.models["current"])#書き出し中に書き換わらないように深いコピーしたいけどできない
                
                try:#dictが書き換わるエラー対策
                    
                    #------------------------------------------
                    info_cr_lso += "-- -- -- -- Dump CurrentModel -- -- -- -- -- \n"   #環世界
                    for lso in currentModel.LSolids():
                        info_cr_lso += "  --- about " + lso.name +" ---  \n"
                        attrs = lso.EnumAttr()
                        keys = sorted( [k for k in attrs.keys()] )
                        
                        position = None#for visualize
                        
                        for key in keys:
                            try:
                                info_cr_lso += str(key) + ":" + str(lso.GetAttr(key)) + "\n"
                                if(key == "position"):
                                    position = lso.GetAttr(key)[0]
                                
                            except Database.Error:
                                info_cr_lso += str(key) + ": not found" +"\n"
                        
                        #visualize##################
                        
                        if cr_con.core_object.spr_creature_visualize_enabled == 1 and\
                            not (position is None):
                            obj = None
                            obj = self.prepare_object("l_" + str(id) + "_" + lso.name.lstrip("so"),id)
                            
                            #パラメータ適用
                            obj.location = (position.x, position.y, position.z)
                            scale = scale_base
                            obj.scale = (scale, scale, scale)
                        
                    #---------------------------------------------
                    info_cr_attv += "attvList \n"  #視覚注意
                    for lso,attv in currentModel.attvList:
                        info_cr_attv += str(lso.name) + ":" + str(attv) +"\n"
                        pos = lso.phSolid.GetPose().getPos()
                        info_cr_attv += "Pos: " + str(pos) + "\n"
                        #落ちる info.append("Ori: " + lso.phSolid.GetPose().getOri())
                        
                        #visualize##################
                        if cr_con.core_object.spr_creature_visualize_enabled == 1:
                            obj = None
                            obj = self.prepare_object("v_" + str(id) + "_" + lso.name.lstrip("so"),id)
                            
                            #パラメータ適用
                            obj.location = (pos.x,pos.y,pos.z)
                            scale = scale_base * attv * scale_attv_factor
                            if scale > scale_max:
                                scale = scale_max
                            obj.scale = (scale, scale, scale)
                        
                    
                    #-------------------------------------------------
                    info_cr_attt += "atttList \n"     #触覚注意
                    for lso,attt in currentModel.atttList:
                        info_cr_attt += str(lso.name) + ":" + str(attt) + "\n"
                        pos = lso.phSolid.GetPose().getPos()
                        info_cr_attt += "Pos: " + str(pos) + "\n"
                        
                        #visualize##################
                        if cr_con.core_object.spr_creature_visualize_enabled == 1:
                            obj = None
                            obj = self.prepare_object("t_" + str(id) + "_" + lso.name.lstrip("so"),id)
                            
                            #パラメータ適用
                            obj.location = (pos.x,pos.y,pos.z)
                            scale = scale_base * attt * scale_attt_factor
                            if scale > scale_max:
                                scale = scale_max
                            obj.scale = (scale, scale, scale)
                    
                    
                    #----------------------------------------------
                    info_cr_max += "Max Solid \n"     #MAX
                    #maxAttSo
                    lso = currentModel.maxAttSo
                    if lso:
                        info_cr_max += "maxAtt: " +  str(lso.name) + "\n"
                        pos = lso.phSolid.GetPose().getPos()
                        
                        if cr_con.core_object.spr_creature_visualize_enabled == 1:
                            obj = self.prepare_object("maxAtt_"+ str(id),id)
                            
                            #パラメータ適用
                            obj.location = (pos.x,pos.y,pos.z)
                            scale = scale_base
                            obj.scale = (scale, scale, scale)
                    
                    
                    #maxAttvSo
                    lso = currentModel.maxAttvSo
                    if lso:
                        info_cr_max +="maxAttv: " + str(lso.name) + "\n"
                        pos = lso.phSolid.GetPose().getPos()
                        
                        if cr_con.core_object.spr_creature_visualize_enabled == 1:
                            obj = self.prepare_object("maxAttv_"+ str(id),id)
                            
                            #パラメータ適用
                            obj.location = (pos.x,pos.y,pos.z)
                            scale = scale_base
                            obj.scale = (scale, scale, scale)
                    
                    #maxAtttSo
                    lso = currentModel.maxAtttSo
                    if lso:
                        info_cr_max +="maxAttt: " + str(lso.name) + "\n"
                        pos = lso.phSolid.GetPose().getPos()
                        
                        if cr_con.core_object.spr_creature_visualize_enabled == 1:
                            obj = self.prepare_object("maxAttt_"+ str(id),id)
                            
                            #パラメータ適用
                            obj.location = (pos.x,pos.y,pos.z)
                            scale = scale_base
                            obj.scale = (scale, scale, scale)
                    
                    
                    
                    #info書き込み-----------------------------------------
                    self.ph_con.info["cr_" + str(id) +"_lso"] = info_cr_lso
                    self.ph_con.info["cr_" + str(id) +"_attv"] = info_cr_attv
                    self.ph_con.info["cr_" + str(id) +"_attt"] = info_cr_attt
                    self.ph_con.info["cr_" + str(id) +"_max"] = info_cr_max
                
                except Exception as e:
                    print("===Error===")
                    print("type: " + str(type(e)) )
                    print("args: " + str(e.args) )
                    #message無い print("message:" + str(e.message) )
                    print("about Exception: " + str(e) )
                    print("======")

    
    #Debug Out
    def update_debug_out(self):
        
        for key, info in self.ph_con.info.items():
            try:
                spr_text = bpy.data.texts["spr_" + key]
            except:
                spr_text = bpy.data.texts.new("spr_" + key)
            
            spr_text.from_string(info)
    

        
    def update_view_angle(self):
        
        context_obj = bpy.context.object
        
        
        obj_heads = {}
        obj_cores = {}
        for obj in bpy.data.objects:
            if obj.spr_CRIKSolid_label == "Head":
                obj_heads[obj.spr_creature_id] = obj
            elif obj.spr_creature_core == 1:
                obj_cores[obj.spr_creature_id] = obj
                
        for id,obj_head in obj_heads.items():
            try:
                obj_core = obj_cores[id]
                view_angle_alt = obj_core.spr_creature_view_angle_alt
                view_angle_az = obj_core.spr_creature_view_angle_az
            except:
                view_angle_alt = 45
                view_angle_az = 60
            
            if (self.prev_view_angle_alt != view_angle_alt or self.prev_view_angle_az != view_angle_az) \
                and obj_cores[id].spr_creature_visualize_enabled == 1:
                
                #Delete Cone######################
                try:
                    view_cone = bpy.data.objects["spr_view_" + str(id)]
                    bpy.ops.object.select_all(action='DESELECT')
                    view_cone.hide_select = False #先にFalseにしないと選択無効 忘れずに
                    view_cone.select = True
                    bpy.ops.object.delete()
                except:
                    view_cone = None
                #Create Cone######################
                depth = 10.0
                vertices_num = 32
                radius_alt = tan(radians(view_angle_alt)) * depth
                radius_az = tan(radians(view_angle_az)) * depth
                
                vertices = [[0 for i in range(0, 3, 1)] for j in range(0, vertices_num + 2, 1)]#2D array {[0,0,0],[0,0,0],...}
                #debug
                #print(cone_verts)
                vertices[0]= (0,0,0)
                i = 1
                while i <= vertices_num:
                    vertices[i][0] = radius_az * cos(i * pi/(vertices_num/2))
                    vertices[i][1] = radius_alt * sin(i * pi/(vertices_num/2))
                    vertices[i][2] = -depth
                    #print(vertices[i])
                    i += 1
                
                vertices[vertices_num + 1] = (0, 0, -depth)
                
                i = 0
                faces =  [[0 for i in range(0, 3, 1)] for j in range(0, vertices_num * 2, 1)]
                while i < (vertices_num - 1):
                    faces[i] = (0, i + 1, i + 2)
                    faces[vertices_num + i] = (i + 2, i + 1, vertices_num + 1)
                    i += 1
                
                faces[vertices_num - 1] = (0, vertices_num, 1)
                faces[vertices_num * 2 - 1] = (1, vertices_num, vertices_num + 1)
                
                view_cone = bl_util.create_mesh("spr_view_" + str(id), (0,0,0), vertices, [], faces)
                
                view_cone.rotation_euler = (radians(90.0), 0.0, 0.0)
                
                #描画関連
                bpy.ops.object.select_all(action='DESELECT')
                view_cone.select = True
                bpy.ops.object.shade_smooth()
                
                view_cone.show_x_ray = True#Trueにするとbuffer shadowがでなくなる・・・
                view_cone.hide_select = True
                
                #Set Material######################
                view_mat = self.preapare_material("spr_view")
                bl_util.set_material(view_cone, view_mat)
                
                #頭にくっつける（オブジェクト中心の座標が一致する）
                view_cone.parent = obj_head
                
                view_cone.spr_creature_id = id
                
                #変動判定用
                self.prev_view_angle_alt = view_angle_alt
                self.prev_view_angle_az = view_angle_az
                
                #Group visualize_obj
                if not "visualize_obj_"+str(id) in bpy.data.groups:
                    bpy.ops.group.create(name="visualize_obj_"+str(id))
                if not obj.name in bpy.data.groups["visualize_obj_"+str(id)].objects:
                    bpy.ops.object.group_link(group="visualize_obj_"+str(id))
                
                #選択対象をもとに戻す
                if(not context_obj is None):
                    bpy.ops.object.select_all(action='DESELECT')
                    context_obj.select = True
                    bpy.context.scene.objects.active = context_obj
                    
    def cr_visualize_display_switch(self,id,switch):
        if "visualize_obj_"+str(id) in bpy.data.groups:
            group = bpy.data.groups["visualize_obj_"+str(id)]
            for name,obj in group.objects.items():
                obj.hide = switch
                obj.hide_render = switch
        #
        #for obj in self.visualize_obj_list:
        #    if(obj.spr_creature_id==id):
        #        obj.hide = switch
        #        obj.hide_render = switch
        #try:
        #    view_cone = bpy.data.objects["spr_view_" + str(id)]
        #    view_cone.hide = switch
        #    view_cone.hide_render = switch
        #    
        #except:
        #    view_cone = None

        
    def prepare_object(self,name,id):
        
        context_obj = bpy.context.object
        
        parameters = {"attv" : "spr_attv",
                    "Mattv" : "spr_attv",
                    "attt" : "spr_attt",
                    "Mattt" : "spr_attt",
                    "lso" : "spr_lso",
                    "Matt" : "spr_lso"}
        
        type = "none"
        
        if "v_" == name[0:2]:#Attv [0:2]->先頭２文字
            type = "attv"
        elif "maxAttv" == name[0:7]:#maxAttv
            type = "Mattv"
        elif "t_" == name[0:2]:#Attt
            type = "attt"
        elif "maxAttt" == name[0:7]:#maxAttt
            type = "Mattt"
        elif "l_" == name[0:2]: #lso
            type = "lso"
        elif "maxAtt" == name[0:6]:#maxAtt
            type = "Matt"
        
        try:
            obj = bpy.data.objects[name]
        except:
            samples = {"attv" : "v_sample_",
                    "Mattv" : "maxAttv_sample_",
                    "attt" : "t_sample_",
                    "Mattt" : "maxAttt_sample_",
                    "lso" : "l_sample_",
                    "Matt" : "maxAtt_sample_"}
            
            if samples[type]+str(id) in bpy.data.objects:   #sampleが存在する場合はそのsampleを使う
                bpy.ops.object.add_named(linked=True, name = samples[type]+str(id) )
                obj = bpy.context.object
                
            
            else:#sampleが存在しない場合は作成する
                obj = None
                
                #オブジェクト生成
                if type == "Mattv" or \
                  type == "Mattt" or \
                  type == "Matt" :
                    bpy.ops.mesh.primitive_torus_add()
                    if type == "Matt" :
                        bpy.context.object.rotation_mode = "QUATERNION"
                        from mathutils import Vector,Quaternion
                        bpy.context.object.rotation_quaternion = Quaternion((0.707, -0.707, 0.0, 0.0))
                    
                else:
                    bpy.ops.mesh.primitive_uv_sphere_add()
                
                obj_sample = bpy.context.object
                obj_sample.name = samples[type]+str(id)
                obj_sample.location = (0, 0, -10)
                obj_sample.spr_creature_id = id
                bpy.ops.object.shade_smooth()
                print(obj_sample.name)
                
                #Set Material######################
                mat = self.preapare_material(parameters[type])
                
                bl_util.set_material(obj_sample, mat)
                
                #linkedオブジェクト作成
                bpy.ops.object.add_named(linked=True, name=samples[type]+str(id) )
                obj = bpy.context.object
            
            
            #命名、設定
            obj.name = name
            obj.show_x_ray = True#Trueにするとシーンのbuffer shadowがでなくなる
            obj.show_name = True
            obj.hide_select = True
            bpy.ops.object.shade_smooth()
            obj.spr_creature_id = id
            
            #Group visualize_obj
            if not "visualize_obj_"+str(id) in bpy.data.groups:
                bpy.ops.group.create(name="visualize_obj_"+str(id))
            if not obj.name in bpy.data.groups["visualize_obj_"+str(id)].objects:
                bpy.ops.object.group_link(group="visualize_obj_"+str(id))
            
            
            if(not context_obj is None):
                bpy.ops.object.select_all(action='DESELECT')
                context_obj.select = True
                bpy.context.scene.objects.active = context_obj
        
        #except ここまで
        
        
        return obj
        
        
    def preapare_material(self,name):
        
        parameters = {"spr_attv":[(0,1,1), (0,1,1), 0.5],   #material name: diffuse, specular, alpha
                    "spr_attt":[(1,1,0), (1,1,0), 0.5],
                    "spr_lso":[(1,0,1), (1,0,1), 0.3],
                    "spr_view":[(0,1,0), (0.5,0.5,0), 0.2]}
        
        try:
            mat = bpy.data.materials[name]
        except:
            mat = None
            
            #material作成
            if name in parameters:
                mat = bl_util.make_material(name, parameters[name][0], parameters[name][1], parameters[name][2]) #name, diffuse, specular, alpha
            else:
                mat = bl_util.make_material(name, (0.7,0.5,0), (0.7,0.5,0), 0.5)
            mat.use_transparency = 1
            mat.use_shadeless = True
            mat.use_cast_buffer_shadows = False
        
        return mat
        


    